@@ -1,7 +1,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2015 The OpenLDAP Foundation.
+ * Copyright 1998-2022 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -45,6 +45,7 @@
 #include <ac/stdlib.h>
 
 #include <ac/errno.h>
+#include <ac/param.h>
 #include <ac/socket.h>
 #include <ac/string.h>
 #include <ac/time.h>
@@ -122,7 +123,7 @@ ldap_send_initial_request(
 	int rc = 1;
 	ber_socket_t sd = AC_SOCKET_INVALID;
 
-	Debug( LDAP_DEBUG_TRACE, "ldap_send_initial_request\n", 0, 0, 0 );
+	Debug0( LDAP_DEBUG_TRACE, "ldap_send_initial_request\n" );
 
 	LDAP_MUTEX_LOCK( &ld->ld_conn_mutex );
 	if ( ber_sockbuf_ctrl( ld->ld_sb, LBER_SB_OPT_GET_FD, &sd ) == -1 ) {
@@ -140,9 +141,8 @@ ldap_send_initial_request(
 		LDAP_MUTEX_UNLOCK( &ld->ld_conn_mutex );
 		return( -1 );
 	} else if ( rc == 0 ) {
-		Debug( LDAP_DEBUG_TRACE,
-			"ldap_open_defconn: successful\n",
-			0, 0, 0 );
+		Debug0( LDAP_DEBUG_TRACE,
+			"ldap_open_defconn: successful\n" );
 	}
 
 #ifdef LDAP_CONNECTIONLESS
@@ -184,7 +184,8 @@ ldap_int_flush_request(
 
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_conn_mutex );
 	if ( ber_flush2( lc->lconn_sb, lr->lr_ber, LBER_FLUSH_FREE_NEVER ) != 0 ) {
-		if ( sock_errno() == EAGAIN ) {
+		if (( sock_errno() == EAGAIN ) || ( sock_errno() == ENOTCONN )) {
+			/* ENOTCONN is returned in Solaris 10 */
 			/* need to continue write later */
 			lr->lr_status = LDAP_REQST_WRITING;
 			ldap_mark_select_write( ld, lc->lconn_sb );
@@ -233,7 +234,7 @@ ldap_send_server_request(
 	int		incparent, rc;
 
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_req_mutex );
-	Debug( LDAP_DEBUG_TRACE, "ldap_send_server_request\n", 0, 0, 0 );
+	Debug0( LDAP_DEBUG_TRACE, "ldap_send_server_request\n" );
 
 	incparent = 0;
 	ld->ld_errno = LDAP_SUCCESS;	/* optimistic */
@@ -315,6 +316,7 @@ ldap_send_server_request(
 		LDAP_MUTEX_UNLOCK( &ld->ld_options.ldo_mutex );
 		if ( rc == -1 ) {
 			ld->ld_errno = LDAP_ENCODING_ERROR;
+			ber_free( ber, 1 );
 			LDAP_CONN_UNLOCK_IF(m_noconn);
 			return rc;
 		}
@@ -327,13 +329,19 @@ ldap_send_server_request(
 	 * request to be in WRITING state.
 	 */
 	rc = 0;
-	if ( ld->ld_requests &&
-		ld->ld_requests->lr_status == LDAP_REQST_WRITING &&
-		ldap_int_flush_request( ld, ld->ld_requests ) < 0 )
-	{
-		rc = -1;
+	if ( ld->ld_requests != NULL ) {
+		TAvlnode *node = ldap_tavl_end( ld->ld_requests, TAVL_DIR_RIGHT );
+		LDAPRequest *lr;
+
+		assert( node != NULL );
+		lr = node->avl_data;
+		if ( lr->lr_status == LDAP_REQST_WRITING &&
+				ldap_int_flush_request( ld, lr ) < 0 ) {
+			rc = -1;
+		}
 	}
 	if ( rc ) {
+		ber_free( ber, 1 );
 		LDAP_CONN_UNLOCK_IF(m_noconn);
 		return rc;
 	}
@@ -397,12 +405,8 @@ ldap_send_server_request(
 		}
 	}
 
-	lr->lr_prev = NULL;
-	lr->lr_next = ld->ld_requests;
-	if ( lr->lr_next != NULL ) {
-		lr->lr_next->lr_prev = lr;
-	}
-	ld->ld_requests = lr;
+	rc = ldap_tavl_insert( &ld->ld_requests, lr, ldap_req_cmp, ldap_avl_dup_error );
+	assert( rc == LDAP_SUCCESS );
 
 	ld->ld_errno = LDAP_SUCCESS;
 	if ( ldap_int_flush_request( ld, lr ) == -1 ) {
@@ -451,7 +455,7 @@ ldap_new_connection( LDAP *ld, LDAPURLDesc **srvlist, int use_ldsb,
 	int		async = 0;
 
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_conn_mutex );
-	Debug( LDAP_DEBUG_TRACE, "ldap_new_connection %d %d %d\n",
+	Debug3( LDAP_DEBUG_TRACE, "ldap_new_connection %d %d %d\n",
 		use_ldsb, connect, (bind != NULL) );
 	/*
 	 * make a new LDAP server connection
@@ -575,7 +579,7 @@ ldap_new_connection( LDAP *ld, LDAPURLDesc **srvlist, int use_ldsb,
 				++lc->lconn_refcnt;	/* avoid premature free */
 				ld->ld_defconn = lc;
 
-				Debug( LDAP_DEBUG_TRACE, "Call application rebind_proc\n", 0, 0, 0);
+				Debug0( LDAP_DEBUG_TRACE, "Call application rebind_proc\n" );
 				LDAP_REQ_UNLOCK_IF(m_req);
 				LDAP_MUTEX_UNLOCK( &ld->ld_conn_mutex );
 				LDAP_RES_UNLOCK_IF(m_res);
@@ -605,9 +609,8 @@ ldap_new_connection( LDAP *ld, LDAPURLDesc **srvlist, int use_ldsb,
 			++lc->lconn_refcnt;	/* avoid premature free */
 			ld->ld_defconn = lc;
 
-			Debug( LDAP_DEBUG_TRACE,
-				"anonymous rebind via ldap_sasl_bind(\"\")\n",
-				0, 0, 0);
+			Debug0( LDAP_DEBUG_TRACE,
+				"anonymous rebind via ldap_sasl_bind(\"\")\n" );
 
 			LDAP_REQ_UNLOCK_IF(m_req);
 			LDAP_MUTEX_UNLOCK( &ld->ld_conn_mutex );
@@ -645,7 +648,7 @@ ldap_new_connection( LDAP *ld, LDAPURLDesc **srvlist, int use_ldsb,
 						break;
 
 					default:
-						Debug( LDAP_DEBUG_TRACE,
+						Debug3( LDAP_DEBUG_TRACE,
 							"ldap_new_connection %p: "
 							"unexpected response %d "
 							"from BIND request id=%d\n",
@@ -732,9 +735,9 @@ ldap_free_connection( LDAP *ld, LDAPConn *lc, int force, int unbind )
 	LDAPConn	*tmplc, *prevlc;
 
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_conn_mutex );
-	Debug( LDAP_DEBUG_TRACE,
+	Debug2( LDAP_DEBUG_TRACE,
 		"ldap_free_connection %d %d\n",
-		force, unbind, 0 );
+		force, unbind );
 
 	if ( force || --lc->lconn_refcnt <= 0 ) {
 		/* remove from connections list first */
@@ -796,28 +799,16 @@ ldap_free_connection( LDAP *ld, LDAPConn *lc, int force, int unbind )
 		}
 
 		ldap_int_sasl_close( ld, lc );
-#ifdef HAVE_GSSAPI
-		ldap_int_gssapi_close( ld, lc );
-#endif
 
 		ldap_free_urllist( lc->lconn_server );
 
 		/* FIXME: is this at all possible?
 		 * ldap_ld_free() in unbind.c calls ldap_free_connection()
 		 * with force == 1 __after__ explicitly calling
-		 * ldap_free_request() on all requests */
+		 * ldap_tavl_free on ld->ld_requests */
 		if ( force ) {
-			LDAPRequest	*lr;
-
-			for ( lr = ld->ld_requests; lr; ) {
-				LDAPRequest	*lr_next = lr->lr_next;
-
-				if ( lr->lr_conn == lc ) {
-					ldap_free_request_int( ld, lr );
-				}
-
-				lr = lr_next;
-			}
+			ldap_tavl_free( ld->ld_requests, ldap_do_free_request );
+			ld->ld_requests = NULL;
 		}
 
 		if ( lc->lconn_sb != ld->ld_sb ) {
@@ -836,14 +827,13 @@ ldap_free_connection( LDAP *ld, LDAPConn *lc, int force, int unbind )
 
 		LDAP_FREE( lc );
 
-		Debug( LDAP_DEBUG_TRACE,
-			"ldap_free_connection: actually freed\n",
-			0, 0, 0 );
+		Debug0( LDAP_DEBUG_TRACE,
+			"ldap_free_connection: actually freed\n" );
 
 	} else {
 		lc->lconn_lastused = time( NULL );
-		Debug( LDAP_DEBUG_TRACE, "ldap_free_connection: refcnt %d\n",
-				lc->lconn_refcnt, 0, 0 );
+		Debug1( LDAP_DEBUG_TRACE, "ldap_free_connection: refcnt %d\n",
+				lc->lconn_refcnt );
 	}
 }
 
@@ -856,24 +846,38 @@ ldap_dump_connection( LDAP *ld, LDAPConn *lconns, int all )
 	LDAPConn	*lc;
    	char		timebuf[32];
 
-	Debug( LDAP_DEBUG_TRACE, "** ld %p Connection%s:\n", (void *)ld, all ? "s" : "", 0 );
+	Debug2( LDAP_DEBUG_TRACE, "** ld %p Connection%s:\n", (void *)ld, all ? "s" : "" );
 	LDAP_MUTEX_LOCK( &ld->ld_conn_mutex );
 	for ( lc = lconns; lc != NULL; lc = lc->lconn_next ) {
 		if ( lc->lconn_server != NULL ) {
-			Debug( LDAP_DEBUG_TRACE, "* host: %s  port: %d%s\n",
+			Debug3( LDAP_DEBUG_TRACE, "* host: %s  port: %d%s\n",
 				( lc->lconn_server->lud_host == NULL ) ? "(null)"
 				: lc->lconn_server->lud_host,
 				lc->lconn_server->lud_port, ( lc->lconn_sb ==
 				ld->ld_sb ) ? "  (default)" : "" );
 		}
-		Debug( LDAP_DEBUG_TRACE, "  refcnt: %d  status: %s\n", lc->lconn_refcnt,
+		if ( lc->lconn_sb != NULL ) {
+			char 			from[LDAP_IPADDRLEN];
+			struct berval 	frombv = BER_BVC(from);
+			ber_socket_t 	sb;
+			if ( ber_sockbuf_ctrl( lc->lconn_sb, LBER_SB_OPT_GET_FD, &sb ) == 1 ) {
+				Sockaddr sin;
+				socklen_t len = sizeof( sin );
+				if ( getsockname( sb, (struct sockaddr *)&sin, &len ) == 0 ) {
+					ldap_pvt_sockaddrstr( &sin, &frombv );
+					Debug1( LDAP_DEBUG_TRACE, "* from: %s\n",
+						( from == NULL ) ? "(null)" : from );
+				}
+			}
+		}
+		Debug2( LDAP_DEBUG_TRACE, "  refcnt: %d  status: %s\n", lc->lconn_refcnt,
 			( lc->lconn_status == LDAP_CONNST_NEEDSOCKET )
 				? "NeedSocket" :
 				( lc->lconn_status == LDAP_CONNST_CONNECTING )
-					? "Connecting" : "Connected", 0 );
-		Debug( LDAP_DEBUG_TRACE, "  last used: %s%s\n",
+					? "Connecting" : "Connected" );
+		Debug2( LDAP_DEBUG_TRACE, "  last used: %s%s\n",
 			ldap_pvt_ctime( &lc->lconn_lastused, timebuf ),
-			lc->lconn_rebind_inprogress ? "  rebind in progress" : "", 0 );
+			lc->lconn_rebind_inprogress ? "  rebind in progress" : "" );
 		if ( lc->lconn_rebind_inprogress ) {
 			if ( lc->lconn_rebind_queue != NULL) {
 				int	i;
@@ -881,15 +885,15 @@ ldap_dump_connection( LDAP *ld, LDAPConn *lconns, int all )
 				for ( i = 0; lc->lconn_rebind_queue[i] != NULL; i++ ) {
 					int	j;
 					for( j = 0; lc->lconn_rebind_queue[i][j] != 0; j++ ) {
-						Debug( LDAP_DEBUG_TRACE, "    queue %d entry %d - %s\n",
+						Debug3( LDAP_DEBUG_TRACE, "    queue %d entry %d - %s\n",
 							i, j, lc->lconn_rebind_queue[i][j] );
 					}
 				}
 			} else {
-				Debug( LDAP_DEBUG_TRACE, "    queue is empty\n", 0, 0, 0 );
+				Debug0( LDAP_DEBUG_TRACE, "    queue is empty\n" );
 			}
 		}
-		Debug( LDAP_DEBUG_TRACE, "\n", 0, 0, 0 );
+		Debug0( LDAP_DEBUG_TRACE, "\n" );
 		if ( !all ) {
 			break;
 		}
@@ -902,18 +906,20 @@ ldap_dump_connection( LDAP *ld, LDAPConn *lconns, int all )
 void
 ldap_dump_requests_and_responses( LDAP *ld )
 {
-	LDAPRequest	*lr;
 	LDAPMessage	*lm, *l;
+	TAvlnode *node;
 	int		i;
 
-	Debug( LDAP_DEBUG_TRACE, "** ld %p Outstanding Requests:\n",
-		(void *)ld, 0, 0 );
-	lr = ld->ld_requests;
-	if ( lr == NULL ) {
-		Debug( LDAP_DEBUG_TRACE, "   Empty\n", 0, 0, 0 );
+	Debug1( LDAP_DEBUG_TRACE, "** ld %p Outstanding Requests:\n",
+		(void *)ld );
+	node = ldap_tavl_end( ld->ld_requests, TAVL_DIR_LEFT );
+	if ( node == NULL ) {
+		Debug0( LDAP_DEBUG_TRACE, "   Empty\n" );
 	}
-	for ( i = 0; lr != NULL; lr = lr->lr_next, i++ ) {
-		Debug( LDAP_DEBUG_TRACE, " * msgid %d,  origid %d, status %s\n",
+	for ( i = 0 ; node != NULL; i++, node = ldap_tavl_next( node, TAVL_DIR_RIGHT ) ) {
+		LDAPRequest	*lr = node->avl_data;
+
+		Debug3( LDAP_DEBUG_TRACE, " * msgid %d,  origid %d, status %s\n",
 			lr->lr_msgid, lr->lr_origid,
 			( lr->lr_status == LDAP_REQST_INPROGRESS ) ? "InProgress" :
 			( lr->lr_status == LDAP_REQST_CHASINGREFS ) ? "ChasingRefs" :
@@ -921,66 +927,48 @@ ldap_dump_requests_and_responses( LDAP *ld )
 			( lr->lr_status == LDAP_REQST_WRITING ) ? "Writing" :
 			( lr->lr_status == LDAP_REQST_COMPLETED ) ? "RequestCompleted"
 				: "InvalidStatus" );
-		Debug( LDAP_DEBUG_TRACE, "   outstanding referrals %d, parent count %d\n",
-			lr->lr_outrefcnt, lr->lr_parentcnt, 0 );
+		Debug2( LDAP_DEBUG_TRACE, "   outstanding referrals %d, parent count %d\n",
+			lr->lr_outrefcnt, lr->lr_parentcnt );
 	}
-	Debug( LDAP_DEBUG_TRACE, "  ld %p request count %d (abandoned %lu)\n",
+	Debug3( LDAP_DEBUG_TRACE, "  ld %p request count %d (abandoned %lu)\n",
 		(void *)ld, i, ld->ld_nabandoned );
-	Debug( LDAP_DEBUG_TRACE, "** ld %p Response Queue:\n", (void *)ld, 0, 0 );
+	Debug1( LDAP_DEBUG_TRACE, "** ld %p Response Queue:\n", (void *)ld );
 	if ( ( lm = ld->ld_responses ) == NULL ) {
-		Debug( LDAP_DEBUG_TRACE, "   Empty\n", 0, 0, 0 );
+		Debug0( LDAP_DEBUG_TRACE, "   Empty\n" );
 	}
 	for ( i = 0; lm != NULL; lm = lm->lm_next, i++ ) {
-		Debug( LDAP_DEBUG_TRACE, " * msgid %d,  type %lu\n",
-		    lm->lm_msgid, (unsigned long)lm->lm_msgtype, 0 );
+		Debug2( LDAP_DEBUG_TRACE, " * msgid %d,  type %lu\n",
+		    lm->lm_msgid, (unsigned long)lm->lm_msgtype );
 		if ( lm->lm_chain != NULL ) {
-			Debug( LDAP_DEBUG_TRACE, "   chained responses:\n", 0, 0, 0 );
+			Debug0( LDAP_DEBUG_TRACE, "   chained responses:\n" );
 			for ( l = lm->lm_chain; l != NULL; l = l->lm_chain ) {
-				Debug( LDAP_DEBUG_TRACE,
+				Debug2( LDAP_DEBUG_TRACE,
 					"  * msgid %d,  type %lu\n",
 					l->lm_msgid,
-					(unsigned long)l->lm_msgtype, 0 );
+					(unsigned long)l->lm_msgtype );
 			}
 		}
 	}
-	Debug( LDAP_DEBUG_TRACE, "  ld %p response count %d\n", (void *)ld, i, 0 );
+	Debug2( LDAP_DEBUG_TRACE, "  ld %p response count %d\n", (void *)ld, i );
 }
 #endif /* LDAP_DEBUG */
 
 /* protected by req_mutex */
-static void
-ldap_free_request_int( LDAP *ld, LDAPRequest *lr )
+void
+ldap_do_free_request( void *arg )
 {
-	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_req_mutex );
+	LDAPRequest *lr = arg;
+
+	Debug3( LDAP_DEBUG_TRACE, "ldap_do_free_request: "
+			"asked to free lr %p msgid %d refcnt %d\n",
+			lr, lr->lr_msgid, lr->lr_refcnt );
 	/* if lr_refcnt > 0, the request has been looked up 
 	 * by ldap_find_request_by_msgid(); if in the meanwhile
 	 * the request is free()'d by someone else, just decrease
-	 * the reference count and extract it from the request
-	 * list; later on, it will be freed. */
-	if ( lr->lr_prev == NULL ) {
-		if ( lr->lr_refcnt == 0 ) {
-			/* free'ing the first request? */
-			assert( ld->ld_requests == lr );
-		}
-
-		if ( ld->ld_requests == lr ) {
-			ld->ld_requests = lr->lr_next;
-		}
-
-	} else {
-		lr->lr_prev->lr_next = lr->lr_next;
-	}
-
-	if ( lr->lr_next != NULL ) {
-		lr->lr_next->lr_prev = lr->lr_prev;
-	}
-
+	 * the reference count; later on, it will be freed. */
 	if ( lr->lr_refcnt > 0 ) {
+		assert( lr->lr_refcnt == 1 );
 		lr->lr_refcnt = -lr->lr_refcnt;
-
-		lr->lr_prev = NULL;
-		lr->lr_next = NULL;
-
 		return;
 	}
 
@@ -1002,13 +990,36 @@ ldap_free_request_int( LDAP *ld, LDAPRequest *lr )
 	LDAP_FREE( lr );
 }
 
+int
+ldap_req_cmp( const void *l, const void *r )
+{
+	const LDAPRequest *left = l, *right = r;
+	return left->lr_msgid - right->lr_msgid;
+}
+
+/* protected by req_mutex */
+static void
+ldap_free_request_int( LDAP *ld, LDAPRequest *lr )
+{
+	LDAPRequest *removed;
+
+	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_req_mutex );
+	removed = ldap_tavl_delete( &ld->ld_requests, lr, ldap_req_cmp );
+	assert( !removed || removed == lr );
+	Debug3( LDAP_DEBUG_TRACE, "ldap_free_request_int: "
+			"lr %p msgid %d%s removed\n",
+			lr, lr->lr_msgid, removed ? "" : " not" );
+
+	ldap_do_free_request( lr );
+}
+
 /* protected by req_mutex */
 void
 ldap_free_request( LDAP *ld, LDAPRequest *lr )
 {
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_req_mutex );
-	Debug( LDAP_DEBUG_TRACE, "ldap_free_request (origid %d, msgid %d)\n",
-		lr->lr_origid, lr->lr_msgid, 0 );
+	Debug2( LDAP_DEBUG_TRACE, "ldap_free_request (origid %d, msgid %d)\n",
+		lr->lr_origid, lr->lr_msgid );
 
 	/* free all referrals (child requests) */
 	while ( lr->lr_child ) {
@@ -1071,7 +1082,7 @@ static int ldap_int_nextref(
  *              The array will be free'd by this function when no longer needed
  *  (IN) sref != 0 if following search reference
  *  (OUT) errstrp = Place to return a string of referrals which could not be followed
- *  (OUT) hadrefp = 1 if sucessfully followed referral
+ *  (OUT) hadrefp = 1 if successfully followed referral
  *
  * Return value - number of referrals followed
  *
@@ -1094,7 +1105,7 @@ ldap_chase_v3referrals( LDAP *ld, LDAPRequest *lr, char **refs, int sref, char *
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_res_mutex );
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_conn_mutex );
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_req_mutex );
-	Debug( LDAP_DEBUG_TRACE, "ldap_chase_v3referrals\n", 0, 0, 0 );
+	Debug0( LDAP_DEBUG_TRACE, "ldap_chase_v3referrals\n" );
 
 	ld->ld_errno = LDAP_SUCCESS;	/* optimistic */
 	*hadrefp = 0;
@@ -1110,8 +1121,8 @@ ldap_chase_v3referrals( LDAP *ld, LDAPRequest *lr, char **refs, int sref, char *
 
 	/* Check for hop limit exceeded */
 	if ( lr->lr_parentcnt >= ld->ld_refhoplimit ) {
-		Debug( LDAP_DEBUG_ANY,
-		    "more than %d referral hops (dropping)\n", ld->ld_refhoplimit, 0, 0 );
+		Debug1( LDAP_DEBUG_ANY,
+		    "more than %d referral hops (dropping)\n", ld->ld_refhoplimit );
 		ld->ld_errno = LDAP_REFERRAL_LIMIT_EXCEEDED;
 		rc = -1;
 		goto done;
@@ -1198,9 +1209,9 @@ ldap_chase_v3referrals( LDAP *ld, LDAPRequest *lr, char **refs, int sref, char *
 				 * if two search references come in one behind the other
 				 * for the same server with different contexts.
 				 */
-				Debug( LDAP_DEBUG_TRACE,
+				Debug1( LDAP_DEBUG_TRACE,
 					"ldap_chase_v3referrals: queue referral \"%s\"\n",
-					refarray[i], 0, 0);
+					refarray[i] );
 				if( lc->lconn_rebind_queue == NULL ) {
 					/* Create a referral list */
 					lc->lconn_rebind_queue =
@@ -1263,9 +1274,9 @@ ldap_chase_v3referrals( LDAP *ld, LDAPRequest *lr, char **refs, int sref, char *
 			goto done;
 		}
 
-		Debug( LDAP_DEBUG_TRACE,
+		Debug2( LDAP_DEBUG_TRACE,
 			"ldap_chase_v3referral: msgid %d, url \"%s\"\n",
-			lr->lr_msgid, refarray[i], 0);
+			lr->lr_msgid, refarray[i] );
 
 		/* Send the new request to the server - may require a bind */
 		rinfo.ri_msgid = origreq->lr_origid;
@@ -1274,7 +1285,7 @@ ldap_chase_v3referrals( LDAP *ld, LDAPRequest *lr, char **refs, int sref, char *
 			origreq, &srv, NULL, &rinfo, 0, 1 );
 		if ( rc < 0 ) {
 			/* Failure, try next referral in the list */
-			Debug( LDAP_DEBUG_ANY, "Unable to chase referral \"%s\" (%d: %s)\n", 
+			Debug3( LDAP_DEBUG_ANY, "Unable to chase referral \"%s\" (%d: %s)\n",
 				refarray[i], ld->ld_errno, ldap_err2string( ld->ld_errno ) );
 			unfollowedcnt += ldap_append_referral( ld, &unfollowed, refarray[i] );
 			ldap_free_urllist( srv );
@@ -1359,7 +1370,7 @@ ldap_chase_referrals( LDAP *ld,
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_res_mutex );
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_conn_mutex );
 	LDAP_ASSERT_MUTEX_OWNER( &ld->ld_req_mutex );
-	Debug( LDAP_DEBUG_TRACE, "ldap_chase_referrals\n", 0, 0, 0 );
+	Debug0( LDAP_DEBUG_TRACE, "ldap_chase_referrals\n" );
 
 	ld->ld_errno = LDAP_SUCCESS;	/* optimistic */
 	*hadrefp = 0;
@@ -1382,9 +1393,9 @@ ldap_chase_referrals( LDAP *ld,
 	}
 
 	if ( lr->lr_parentcnt >= ld->ld_refhoplimit ) {
-		Debug( LDAP_DEBUG_ANY,
+		Debug1( LDAP_DEBUG_ANY,
 		    "more than %d referral hops (dropping)\n",
-		    ld->ld_refhoplimit, 0, 0 );
+		    ld->ld_refhoplimit );
 		    /* XXX report as error in ld->ld_errno? */
 		    return( 0 );
 	}
@@ -1407,16 +1418,16 @@ ldap_chase_referrals( LDAP *ld,
 
 		rc = ldap_url_parse_ext( ref, &srv, LDAP_PVT_URL_PARSE_NOEMPTY_DN );
 		if ( rc != LDAP_URL_SUCCESS ) {
-			Debug( LDAP_DEBUG_TRACE,
+			Debug2( LDAP_DEBUG_TRACE,
 				"ignoring %s referral <%s>\n",
-				ref, rc == LDAP_URL_ERR_BADSCHEME ? "unknown" : "incorrect", 0 );
+				ref, rc == LDAP_URL_ERR_BADSCHEME ? "unknown" : "incorrect" );
 			rc = ldap_append_referral( ld, &unfollowed, ref );
 			*hadrefp = 1;
 			continue;
 		}
 
-		Debug( LDAP_DEBUG_TRACE,
-		    "chasing LDAP referral: <%s>\n", ref, 0, 0 );
+		Debug1( LDAP_DEBUG_TRACE,
+		    "chasing LDAP referral: <%s>\n", ref );
 
 		*hadrefp = 1;
 
@@ -1464,7 +1475,7 @@ ldap_chase_referrals( LDAP *ld,
 		if( rc >= 0 ) {
 			++count;
 		} else {
-			Debug( LDAP_DEBUG_ANY,
+			Debug3( LDAP_DEBUG_ANY,
 				"Unable to chase referral \"%s\" (%d: %s)\n", 
 				ref, ld->ld_errno, ldap_err2string( ld->ld_errno ) );
 			rc = ldap_append_referral( ld, &unfollowed, ref );
@@ -1532,10 +1543,10 @@ re_encode_request( LDAP *ld,
 	BerElement	tmpber, *ber;
 	struct berval		dn;
 
-	Debug( LDAP_DEBUG_TRACE,
+	Debug2( LDAP_DEBUG_TRACE,
 	    "re_encode_request: new msgid %ld, new dn <%s>\n",
 	    (long) msgid,
-		( srv == NULL || srv->lud_dn == NULL) ? "NONE" : srv->lud_dn, 0 );
+		( srv == NULL || srv->lud_dn == NULL) ? "NONE" : srv->lud_dn );
 
 	tmpber = *origber;
 
@@ -1637,8 +1648,7 @@ re_encode_request( LDAP *ld,
 
 #ifdef LDAP_DEBUG
 	if ( ldap_debug & LDAP_DEBUG_PACKETS ) {
-		Debug( LDAP_DEBUG_ANY, "re_encode_request new request is:\n",
-		    0, 0, 0 );
+		Debug0( LDAP_DEBUG_ANY, "re_encode_request new request is:\n" );
 		ber_log_dump( LDAP_DEBUG_BER, ldap_debug, ber, 0 );
 	}
 #endif /* LDAP_DEBUG */
@@ -1652,19 +1662,24 @@ re_encode_request( LDAP *ld,
 LDAPRequest *
 ldap_find_request_by_msgid( LDAP *ld, ber_int_t msgid )
 {
-	LDAPRequest	*lr;
+	LDAPRequest	*lr, needle = {0};
+	needle.lr_msgid = msgid;
 
-	for ( lr = ld->ld_requests; lr != NULL; lr = lr->lr_next ) {
-		if ( lr->lr_status == LDAP_REQST_COMPLETED ) {
-			continue;	/* Skip completed requests */
-		}
-		if ( msgid == lr->lr_msgid ) {
-			lr->lr_refcnt++;
-			break;
-		}
+	lr = ldap_tavl_find( ld->ld_requests, &needle, ldap_req_cmp );
+	if ( lr != NULL && lr->lr_status != LDAP_REQST_COMPLETED ) {
+		/* lr_refcnt is only negative when we removed it from ld_requests
+		 * already, it is positive if we have sub-requests (referrals) */
+		assert( lr->lr_refcnt >= 0 );
+		lr->lr_refcnt++;
+		Debug3( LDAP_DEBUG_TRACE, "ldap_find_request_by_msgid: "
+				"msgid %d, lr %p lr->lr_refcnt = %d\n",
+				msgid, lr, lr->lr_refcnt );
+		return lr;
 	}
 
-	return( lr );
+	Debug2( LDAP_DEBUG_TRACE, "ldap_find_request_by_msgid: "
+			"msgid %d, lr %p\n", msgid, lr );
+	return NULL;
 }
 
 /* protected by req_mutex */
@@ -1673,23 +1688,26 @@ ldap_return_request( LDAP *ld, LDAPRequest *lrx, int freeit )
 {
 	LDAPRequest	*lr;
 
-	for ( lr = ld->ld_requests; lr != NULL; lr = lr->lr_next ) {
-		if ( lr == lrx ) {
-			if ( lr->lr_refcnt > 0 ) {
-				lr->lr_refcnt--;
-
-			} else if ( lr->lr_refcnt < 0 ) {
-				lr->lr_refcnt++;
-				if ( lr->lr_refcnt == 0 ) {
-					lr = NULL;
-				}
+	lr = ldap_tavl_find( ld->ld_requests, lrx, ldap_req_cmp );
+	Debug2( LDAP_DEBUG_TRACE, "ldap_return_request: "
+			"lrx %p, lr %p\n", lrx, lr );
+	if ( lr ) {
+		assert( lr == lrx );
+		if ( lr->lr_refcnt > 0 ) {
+			lr->lr_refcnt--;
+		} else if ( lr->lr_refcnt < 0 ) {
+			lr->lr_refcnt++;
+			if ( lr->lr_refcnt == 0 ) {
+				lr = NULL;
 			}
-			break;
 		}
 	}
+	Debug3( LDAP_DEBUG_TRACE, "ldap_return_request: "
+			"lrx->lr_msgid %d, lrx->lr_refcnt is now %d, lr is %s present\n",
+			lrx->lr_msgid, lrx->lr_refcnt, lr ? "still" : "not" );
+	/* The request is not tracked anymore */
 	if ( lr == NULL ) {
 		ldap_free_request_int( ld, lrx );
-
 	} else if ( freeit ) {
 		ldap_free_request( ld, lrx );
 	}
