@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/libraries/liblutil/sasl.c,v 1.22.2.5 2010/04/13 20:23:06 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2010 The OpenLDAP Foundation.
+ * Copyright 1998-2022 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -216,8 +216,6 @@ int lutil_sasl_interact(
 {
 	sasl_interact_t *interact = in;
 
-	if( ld == NULL ) return LDAP_PARAM_ERROR;
-
 	if( flags == LDAP_SASL_INTERACTIVE ) {
 		fputs( _("SASL Interaction\n"), stderr );
 	}
