@@ -1,8 +1,8 @@
 /* aclparse.c - routines to parse and check acl's */
-/* $OpenLDAP: pkg/ldap/servers/slapd/aclparse.c,v 1.198.2.13 2010/04/15 22:26:06 quanah Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2010 The OpenLDAP Foundation.
+ * Copyright 1998-2022 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -57,6 +57,9 @@ const char *style_strings[] = {
 	NULL
 };
 
+#define ACLBUF_CHUNKSIZE	8192
+static struct berval aclbuf;
+
 static void		split(char *line, int splitchar, char **left, char **right);
 static void		access_append(Access **l, Access *a);
 static void		access_free( Access *a );
@@ -165,12 +168,9 @@ regtest(const char *fname, int lineno, char *pat) {
 
 		regerror(e, &re, error, sizeof(error));
 
-		snprintf( buf, sizeof( buf ),
-			"regular expression \"%s\" bad because of %s",
-			pat, error );
-		Debug( LDAP_DEBUG_ANY,
-			"%s: line %d: %s\n",
-			fname, lineno, buf );
+		Debug(LDAP_DEBUG_ANY,
+		      "%s: line %d: regular expression \"%s\" bad because of %s\n",
+		      fname, lineno, pat, error );
 		acl_usage();
 		exit( EXIT_FAILURE );
 	}
@@ -340,7 +340,7 @@ parse_acl(
 			if ( a != NULL ) {
 				Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 					"only one to clause allowed in access line\n",
-				    fname, lineno, 0 );
+				    fname, lineno );
 				goto fail;
 			}
 			a = (AccessControl *) ch_calloc( 1, sizeof(AccessControl) );
@@ -358,7 +358,7 @@ parse_acl(
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: dn pattern"
 							" already specified in to clause.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -383,7 +383,7 @@ parse_acl(
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: dn pattern"
 							" already specified in to clause.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -464,7 +464,7 @@ parse_acl(
 							"%s: line %d: \"attr\" "
 							"is deprecated (and undocumented); "
 							"use \"attrs\" instead.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 					}
 
 					a->acl_attrs = str2anlist( a->acl_attrs,
@@ -483,14 +483,14 @@ parse_acl(
 					if ( !BER_BVISEMPTY( &a->acl_attrval ) ) {
 						Debug( LDAP_DEBUG_ANY,
 				"%s: line %d: attr val already specified in to clause.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 					if ( a->acl_attrs == NULL || !BER_BVISEMPTY( &a->acl_attrs[1].an_name ) )
 					{
 						Debug( LDAP_DEBUG_ANY,
 				"%s: line %d: attr val requires a single attribute.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -512,16 +512,11 @@ parse_acl(
 
 						if( !mr_usable_with_at( a->acl_attrval_mr, a->acl_attrs[ 0 ].an_desc->ad_type ) )
 						{
-							char	buf[ SLAP_TEXT_BUFLEN ];
-
-							snprintf( buf, sizeof( buf ),
-								"matching rule \"%s\" use "
-								"with attr \"%s\" not appropriate.",
-								mr, a->acl_attrs[ 0 ].an_name.bv_val );
-								
-
-							Debug( LDAP_DEBUG_ANY, "%s: line %d: %s\n",
-								fname, lineno, buf );
+							Debug(LDAP_DEBUG_ANY,
+							      "%s: line %d: matching rule \"%s\" use " "with attr \"%s\" not appropriate.\n",
+							      fname, lineno,
+							      mr,
+							      a->acl_attrs[0].an_name.bv_val );
 							goto fail;
 						}
 					}
@@ -531,17 +526,12 @@ parse_acl(
 							int e = regcomp( &a->acl_attrval_re, bv.bv_val,
 								REG_EXTENDED | REG_ICASE );
 							if ( e ) {
-								char	err[SLAP_TEXT_BUFLEN],
-									buf[ SLAP_TEXT_BUFLEN ];
+								char	err[SLAP_TEXT_BUFLEN];
 
 								regerror( e, &a->acl_attrval_re, err, sizeof( err ) );
-
-								snprintf( buf, sizeof( buf ),
-									"regular expression \"%s\" bad because of %s",
-									right, err );
-
-								Debug( LDAP_DEBUG_ANY, "%s: line %d: %s\n",
-									fname, lineno, buf );
+								Debug(LDAP_DEBUG_ANY,
+								      "%s: line %d: regular expression \"%s\" bad because of %s\n",
+								      fname, lineno, right, err );
 								goto fail;
 							}
 							a->acl_attrval_style = ACL_STYLE_REGEX;
@@ -571,45 +561,34 @@ parse_acl(
 								} else if ( !strcasecmp( style, "children" ) ) {
 									a->acl_attrval_style = ACL_STYLE_CHILDREN;
 								} else {
-									char	buf[ SLAP_TEXT_BUFLEN ];
-
-									snprintf( buf, sizeof( buf ),
-										"unknown val.<style> \"%s\" for attributeType \"%s\" "
-											"with DN syntax.",
-										style,
-										a->acl_attrs[0].an_desc->ad_cname.bv_val );
-
-									Debug( LDAP_DEBUG_CONFIG | LDAP_DEBUG_ACL, 
-										"%s: line %d: %s\n",
-										fname, lineno, buf );
+									Debug(LDAP_DEBUG_CONFIG | LDAP_DEBUG_ACL,
+									      "%s: line %d: unknown val.<style> \"%s\" for attributeType \"%s\" " "with DN syntax.\n",
+									      fname,
+									      lineno,
+									      style,
+									      a->acl_attrs[0].an_desc->ad_cname.bv_val );
 									goto fail;
 								}
 
 								rc = dnNormalize( 0, NULL, NULL, &bv, &a->acl_attrval, NULL );
 								if ( rc != LDAP_SUCCESS ) {
-									char	buf[ SLAP_TEXT_BUFLEN ];
-
-									snprintf( buf, sizeof( buf ),
-										"unable to normalize DN \"%s\" "
-										"for attributeType \"%s\" (%d).",
-										bv.bv_val,
-										a->acl_attrs[0].an_desc->ad_cname.bv_val,
-										rc );
-									Debug( LDAP_DEBUG_ANY, 
-										"%s: line %d: %s\n",
-										fname, lineno, buf );
+									Debug(LDAP_DEBUG_ANY,
+									      "%s: line %d: unable to normalize DN \"%s\" " "for attributeType \"%s\" (%d).\n",
+									      fname,
+									      lineno,
+									      bv.bv_val,
+									      a->acl_attrs[0].an_desc->ad_cname.bv_val,
+									      rc );
 									goto fail;
 								}
 
 							} else {
-								char	buf[ SLAP_TEXT_BUFLEN ];
-
-								snprintf( buf, sizeof( buf ),
-									"unknown val.<style> \"%s\" for attributeType \"%s\".",
-									style, a->acl_attrs[0].an_desc->ad_cname.bv_val );
-								Debug( LDAP_DEBUG_CONFIG | LDAP_DEBUG_ACL, 
-									"%s: line %d: %s\n",
-									fname, lineno, buf );
+								Debug(LDAP_DEBUG_CONFIG | LDAP_DEBUG_ACL,
+								      "%s: line %d: unknown val.<style> \"%s\" for attributeType \"%s\".\n",
+								      fname,
+								      lineno,
+								      style,
+								      a->acl_attrs[0].an_desc->ad_cname.bv_val );
 								goto fail;
 							}
 						}
@@ -643,14 +622,12 @@ parse_acl(
 							&text,
 							NULL );
 						if ( rc != LDAP_SUCCESS ) {
-							char	buf[ SLAP_TEXT_BUFLEN ];
-
-							snprintf( buf, sizeof( buf ), "%s: line %d: "
-								" attr \"%s\" normalization failed (%d: %s)",
-								fname, lineno,
-								a->acl_attrs[ 0 ].an_name.bv_val, rc, text );
-							Debug( LDAP_DEBUG_ANY, "%s: line %d: %s.\n",
-								fname, lineno, buf );
+							Debug(LDAP_DEBUG_ANY,
+							      "%s: line %d: %s: line %d: " " attr \"%s\" normalization failed (%d: %s).\n",
+							      fname, lineno,
+							      fname, lineno,
+							      a->acl_attrs[0].an_name.bv_val,
+							      rc, text );
 							goto fail;
 						}
 					}
@@ -690,15 +667,12 @@ parse_acl(
 					int e = regcomp( &a->acl_dn_re, a->acl_dn_pat.bv_val,
 						REG_EXTENDED | REG_ICASE );
 					if ( e ) {
-						char	err[ SLAP_TEXT_BUFLEN ],
-							buf[ SLAP_TEXT_BUFLEN ];
+						char	err[ SLAP_TEXT_BUFLEN ];
 
 						regerror( e, &a->acl_dn_re, err, sizeof( err ) );
-						snprintf( buf, sizeof( buf ),
-							"regular expression \"%s\" bad because of %s",
-							right, err );
-						Debug( LDAP_DEBUG_ANY, "%s: line %d: %s\n",
-							fname, lineno, buf );
+						Debug(LDAP_DEBUG_ANY,
+						      "%s: line %d: regular expression \"%s\" bad because of %s\n",
+						      fname, lineno, right, err );
 						goto fail;
 					}
 				}
@@ -709,7 +683,7 @@ parse_acl(
 			if ( a == NULL ) {
 				Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 					"to clause required before by clause in access line\n",
-					fname, lineno, 0 );
+					fname, lineno );
 				goto fail;
 			}
 
@@ -720,7 +694,7 @@ parse_acl(
 			if ( ++i == argc ) {
 				Debug( LDAP_DEBUG_ANY,
 					"%s: line %d: premature EOL: expecting <who>\n",
-					fname, lineno, 0 );
+					fname, lineno );
 				goto fail;
 			}
 
@@ -751,13 +725,13 @@ parse_acl(
 								Debug( LDAP_DEBUG_ANY,
 									"%s: line %d: premature eol: "
 									"expecting closing '}' in \"level{n}\"\n",
-									fname, lineno, 0 );
+									fname, lineno );
 								goto fail;
 							} else if ( p == style_level ) {
 								Debug( LDAP_DEBUG_ANY,
 									"%s: line %d: empty level "
 									"in \"level{n}\"\n",
-									fname, lineno, 0 );
+									fname, lineno );
 								goto fail;
 							}
 							p[0] = '\0';
@@ -791,7 +765,7 @@ parse_acl(
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: unable to parse level "
 							"in \"level{n}\"\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -810,7 +784,7 @@ parse_acl(
 #ifndef LDAP_PF_INET6
 					Debug( LDAP_DEBUG_ANY,
 						"%s: line %d: IPv6 not supported\n",
-						fname, lineno, 0 );
+						fname, lineno );
 #endif /* ! LDAP_PF_INET6 */
 					sty = ACL_STYLE_IPV6;
 
@@ -820,7 +794,7 @@ parse_acl(
 					Debug( LDAP_DEBUG_CONFIG | LDAP_DEBUG_ACL,
 						"%s: line %d: "
 						"\"path\" style modifier is useless without local.\n",
-						fname, lineno, 0 );
+						fname, lineno );
 					goto fail;
 #endif /* LDAP_PF_LOCAL */
 
@@ -838,7 +812,7 @@ parse_acl(
 					case ACL_STYLE_REGEX:
 						Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 							"\"regex\" style implies \"expand\" modifier.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 						break;
 
@@ -952,7 +926,7 @@ parse_acl(
 					if ( !BER_BVISEMPTY( &bdn->a_pat ) ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: dn pattern already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -979,7 +953,7 @@ parse_acl(
 							Debug( LDAP_DEBUG_ANY,
 								"%s: line %d: rootdn is always granted "
 								"unlimited privileges.\n",
-								fname, lineno, 0 );
+								fname, lineno );
 						}
 
 					} else {
@@ -1008,7 +982,7 @@ parse_acl(
 						} else {
 							Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 								"\"expand\" used with no expansions in \"pattern\".\n",
-								fname, lineno, 0 );
+								fname, lineno );
 							goto fail;
 						} 
 					}
@@ -1026,12 +1000,12 @@ parse_acl(
 							Debug( LDAP_DEBUG_ANY,
 								"%s: line %d: \"onelevel\" should be used "
 								"instead of \"level{1}\" in by DN clause\n",
-								fname, lineno, 0 );
+								fname, lineno );
 						} else if ( level == 0 && sty == ACL_STYLE_LEVEL ) {
 							Debug( LDAP_DEBUG_ANY,
 								"%s: line %d: \"base\" should be used "
 								"instead of \"level{0}\" in by DN clause\n",
-								fname, lineno, 0 );
+								fname, lineno );
 						}
 
 						bdn->a_level = level;
@@ -1051,21 +1025,17 @@ parse_acl(
 					if( bdn->a_at != NULL ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: dnattr already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
 					rc = slap_str2ad( right, &bdn->a_at, &text );
 
 					if( rc != LDAP_SUCCESS ) {
-						char	buf[ SLAP_TEXT_BUFLEN ];
-
-						snprintf( buf, sizeof( buf ),
-							"dnattr \"%s\": %s",
-							right, text );
-						Debug( LDAP_DEBUG_ANY,
-							"%s: line %d: %s\n",
-							fname, lineno, buf );
+						Debug(LDAP_DEBUG_ANY,
+						      "%s: line %d: dnattr \"%s\": %s\n",
+						      fname, lineno, right,
+						      text );
 						goto fail;
 					}
 
@@ -1075,16 +1045,10 @@ parse_acl(
 						!is_at_syntax( bdn->a_at->ad_type,
 						SLAPD_NAMEUID_SYNTAX ))
 					{
-						char	buf[ SLAP_TEXT_BUFLEN ];
-
-						snprintf( buf, sizeof( buf ),
-							"dnattr \"%s\": "
-							"inappropriate syntax: %s\n",
-							right,
-							bdn->a_at->ad_type->sat_syntax_oid );
-						Debug( LDAP_DEBUG_ANY,
-							"%s: line %d: %s\n",
-							fname, lineno, buf );
+						Debug(LDAP_DEBUG_ANY,
+						      "%s: line %d: dnattr \"%s\": " "inappropriate syntax: %s\n\n",
+						      fname, lineno, right,
+						      bdn->a_at->ad_type->sat_syntax_oid );
 						goto fail;
 					}
 
@@ -1111,7 +1075,7 @@ parse_acl(
 							"%s: line %d: "
 							"deprecated group style \"regex\"; "
 							"use \"expand\" instead.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						sty = ACL_STYLE_EXPAND;
 						break;
 
@@ -1142,7 +1106,7 @@ parse_acl(
 					if ( !BER_BVISEMPTY( &b->a_group_pat ) ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: group pattern already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1227,14 +1191,10 @@ parse_acl(
 
 					rc = slap_str2ad( attr_name, &b->a_group_at, &text );
 					if ( rc != LDAP_SUCCESS ) {
-						char	buf[ SLAP_TEXT_BUFLEN ];
-
-						snprintf( buf, sizeof( buf ),
-							"group \"%s\": %s.",
-							right, text );
-						Debug( LDAP_DEBUG_ANY,
-							"%s: line %d: %s\n",
-							fname, lineno, buf );
+						Debug(LDAP_DEBUG_ANY,
+						      "%s: line %d: group \"%s\": %s.\n",
+						      fname, lineno, right,
+						      text );
 						goto fail;
 					}
 
@@ -1245,19 +1205,11 @@ parse_acl(
 						&& !is_at_subtype( b->a_group_at->ad_type,
 							slap_schema.si_ad_labeledURI->ad_type ) /* e.g. memberURL */ )
 					{
-						char	buf[ SLAP_TEXT_BUFLEN ];
-
-						snprintf( buf, sizeof( buf ),
-							"group \"%s\" attr \"%s\": inappropriate syntax: %s; "
-							"must be " SLAPD_DN_SYNTAX " (DN), "
-							SLAPD_NAMEUID_SYNTAX " (NameUID) "
-							"or a subtype of labeledURI.",
-							right,
-							attr_name,
-							at_syntax( b->a_group_at->ad_type ) );
-						Debug( LDAP_DEBUG_ANY,
-							"%s: line %d: %s\n",
-							fname, lineno, buf );
+						Debug(LDAP_DEBUG_ANY,
+						      "%s: line %d: group \"%s\" attr \"%s\": inappropriate syntax: %s; " "must be " SLAPD_DN_SYNTAX " (DN), " SLAPD_NAMEUID_SYNTAX " (NameUID) " "or a subtype of labeledURI.\n",
+						      fname, lineno, right,
+						      attr_name,
+						      at_syntax(b->a_group_at->ad_type) );
 						goto fail;
 					}
 
@@ -1273,14 +1225,11 @@ parse_acl(
 							ocs, NULL );
 
 						if( rc != 0 ) {
-							char	buf[ SLAP_TEXT_BUFLEN ];
-
-							snprintf( buf, sizeof( buf ),
-								"group: \"%s\" not allowed by \"%s\".",
-								b->a_group_at->ad_cname.bv_val,
-								b->a_group_oc->soc_oid );
-							Debug( LDAP_DEBUG_ANY, "%s: line %d: %s\n",
-								fname, lineno, buf );
+							Debug(LDAP_DEBUG_ANY,
+							      "%s: line %d: group: \"%s\" not allowed by \"%s\".\n",
+							      fname, lineno,
+							      b->a_group_at->ad_cname.bv_val,
+							      b->a_group_oc->soc_oid );
 							goto fail;
 						}
 					}
@@ -1318,7 +1267,7 @@ parse_acl(
 					if ( !BER_BVISEMPTY( &b->a_peername_pat ) ) {
 						Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 							"peername pattern already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1458,7 +1407,7 @@ parse_acl(
 					if ( !BER_BVISNULL( &b->a_sockname_pat ) ) {
 						Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 							"sockname pattern already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1491,7 +1440,7 @@ parse_acl(
 								"%s: line %d: "
 								"\"expand\" modifier "
 								"with \"expand\" style.\n",
-								fname, lineno, 0 );
+								fname, lineno );
 						}
 						sty = ACL_STYLE_BASE;
 						expand = 1;
@@ -1516,7 +1465,7 @@ parse_acl(
 					if ( !BER_BVISEMPTY( &b->a_domain_pat ) ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: domain pattern already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1563,7 +1512,7 @@ parse_acl(
 					if ( !BER_BVISEMPTY( &b->a_sockurl_pat ) ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: sockurl pattern already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1590,7 +1539,7 @@ parse_acl(
 							"deprecated set style "
 							"\"regex\" in <by> clause; "
 							"use \"expand\" instead.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						sty = ACL_STYLE_EXPAND;
 						/* FALLTHRU */
 						
@@ -1608,14 +1557,14 @@ parse_acl(
 					if ( !BER_BVISEMPTY( &b->a_set_pat ) ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: set attribute already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
 					if ( right == NULL || *right == '\0' ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: no set is defined.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1635,7 +1584,7 @@ parse_acl(
 						Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 							"undocumented deprecated \"aci\" directive "
 							"is superseded by \"dynacl/aci\".\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						name = "aci";
 						
 					} else
@@ -1673,14 +1622,14 @@ parse_acl(
 					if ( b->a_authz.sai_ssf ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: ssf attribute already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
 					if ( right == NULL || *right == '\0' ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: no ssf is defined.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1711,14 +1660,14 @@ parse_acl(
 					if ( b->a_authz.sai_transport_ssf ) {
 						Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 							"transport_ssf attribute already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
 					if ( right == NULL || *right == '\0' ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: no transport_ssf is defined.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1749,14 +1698,14 @@ parse_acl(
 					if ( b->a_authz.sai_tls_ssf ) {
 						Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 							"tls_ssf attribute already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
 					if ( right == NULL || *right == '\0' ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: no tls_ssf is defined\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1787,14 +1736,14 @@ parse_acl(
 					if ( b->a_authz.sai_sasl_ssf ) {
 						Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 							"sasl_ssf attribute already specified.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
 					if ( right == NULL || *right == '\0' ) {
 						Debug( LDAP_DEBUG_ANY,
 							"%s: line %d: no sasl_ssf is defined.\n",
-							fname, lineno, 0 );
+							fname, lineno );
 						goto fail;
 					}
 
@@ -1925,7 +1874,7 @@ parse_acl(
 	if ( a == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 			"warning: no access clause(s) specified in access line.\n",
-			fname, lineno, 0 );
+			fname, lineno );
 		goto fail;
 
 	} else {
@@ -1938,7 +1887,7 @@ parse_acl(
 		if ( a->acl_access == NULL ) {
 			Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 				"warning: no by clause(s) specified in access line.\n",
-				fname, lineno, 0 );
+				fname, lineno );
 			goto fail;
 		}
 
@@ -1946,13 +1895,13 @@ parse_acl(
 			if ( be->be_nsuffix == NULL ) {
 				Debug( LDAP_DEBUG_ACL, "%s: line %d: warning: "
 					"scope checking needs suffix before ACLs.\n",
-					fname, lineno, 0 );
+					fname, lineno );
 				/* go ahead, since checking is not authoritative */
 			} else if ( !BER_BVISNULL( &be->be_nsuffix[ 1 ] ) ) {
 				Debug( LDAP_DEBUG_ACL, "%s: line %d: warning: "
 					"scope checking only applies to single-valued "
 					"suffix databases\n",
-					fname, lineno, 0 );
+					fname, lineno );
 				/* go ahead, since checking is not authoritative */
 			} else {
 				switch ( check_scope( be, a ) ) {
@@ -1960,27 +1909,27 @@ parse_acl(
 					Debug( LDAP_DEBUG_ACL, "%s: line %d: warning: "
 						"cannot assess the validity of the ACL scope within "
 						"backend naming context\n",
-						fname, lineno, 0 );
+						fname, lineno );
 					break;
 
 				case ACL_SCOPE_WARN:
 					Debug( LDAP_DEBUG_ACL, "%s: line %d: warning: "
 						"ACL could be out of scope within backend naming context\n",
-						fname, lineno, 0 );
+						fname, lineno );
 					break;
 
 				case ACL_SCOPE_PARTIAL:
 					Debug( LDAP_DEBUG_ACL, "%s: line %d: warning: "
 						"ACL appears to be partially out of scope within "
 						"backend naming context\n",
-						fname, lineno, 0 );
+						fname, lineno );
 					break;
 	
 				case ACL_SCOPE_ERR:
 					Debug( LDAP_DEBUG_ACL, "%s: line %d: warning: "
 						"ACL appears to be out of scope within "
 						"backend naming context\n",
-						fname, lineno, 0 );
+						fname, lineno );
 					break;
 
 				default:
@@ -2450,6 +2399,11 @@ acl_destroy( AccessControl *a )
 		n = a->acl_next;
 		acl_free( a );
 	}
+
+	if ( !BER_BVISNULL( &aclbuf ) ) {
+		ch_free( aclbuf.bv_val );
+		BER_BVZERO( &aclbuf );
+	}
 }
 
 char *
@@ -2527,9 +2481,39 @@ str2access( const char *str )
 	return( ACL_INVALID_ACCESS );
 }
 
-#define ACLBUF_MAXLEN	8192
+static char *
+safe_strncopy( char *ptr, const char *src, size_t n, struct berval *buf )
+{
+	while ( ptr + n >= buf->bv_val + buf->bv_len ) {
+		char *tmp = ch_realloc( buf->bv_val, 2*buf->bv_len );
+		if ( tmp == NULL ) {
+			return NULL;
+		}
+		ptr = tmp + (ptr - buf->bv_val);
+		buf->bv_val = tmp;
+		buf->bv_len *= 2;
+	}
 
-static char aclbuf[ACLBUF_MAXLEN];
+	return lutil_strncopy( ptr, src, n );
+}
+
+static char *
+safe_strcopy( char *ptr, const char *s, struct berval *buf )
+{
+	size_t n = strlen( s );
+
+	return safe_strncopy( ptr, s, n, buf );
+}
+
+static char *
+safe_strbvcopy( char *ptr, const struct berval *bv, struct berval *buf )
+{
+	return safe_strncopy( ptr, bv->bv_val, bv->bv_len, buf );
+}
+
+#define acl_safe_strcopy( ptr, s ) safe_strcopy( (ptr), (s), &aclbuf )
+#define acl_safe_strncopy( ptr, s, n ) safe_strncopy( (ptr), (s), (n), &aclbuf )
+#define acl_safe_strbvcopy( ptr, bv ) safe_strbvcopy( (ptr), (bv), &aclbuf )
 
 static char *
 dnaccess2text( slap_dn_access *bdn, char *ptr, int is_realdn )
@@ -2537,7 +2521,7 @@ dnaccess2text( slap_dn_access *bdn, char *ptr, int is_realdn )
 	*ptr++ = ' ';
 
 	if ( is_realdn ) {
-		ptr = lutil_strcopy( ptr, "real" );
+		ptr = acl_safe_strcopy( ptr, "real" );
 	}
 
 	if ( ber_bvccmp( &bdn->a_pat, '*' ) ||
@@ -2549,33 +2533,34 @@ dnaccess2text( slap_dn_access *bdn, char *ptr, int is_realdn )
 			assert( ! ber_bvccmp( &bdn->a_pat, '*' ) );
 		}
 			
-		ptr = lutil_strcopy( ptr, bdn->a_pat.bv_val );
+		ptr = acl_safe_strbvcopy( ptr, &bdn->a_pat );
 		if ( bdn->a_style == ACL_STYLE_SELF && bdn->a_self_level != 0 ) {
-			int n = sprintf( ptr, ".level{%d}", bdn->a_self_level );
+			char buf[SLAP_TEXT_BUFLEN];
+			int n = snprintf( buf, sizeof(buf), ".level{%d}", bdn->a_self_level );
 			if ( n > 0 ) {
-				ptr += n;
+				ptr = acl_safe_strncopy( ptr, buf, n );
 			} /* else ? */
 		}
 
 	} else {
-		ptr = lutil_strcopy( ptr, "dn." );
+		ptr = acl_safe_strcopy( ptr, "dn." );
 		if ( bdn->a_style == ACL_STYLE_BASE )
-			ptr = lutil_strcopy( ptr, style_base );
+			ptr = acl_safe_strcopy( ptr, style_base );
 		else 
-			ptr = lutil_strcopy( ptr, style_strings[bdn->a_style] );
+			ptr = acl_safe_strcopy( ptr, style_strings[bdn->a_style] );
 		if ( bdn->a_style == ACL_STYLE_LEVEL ) {
-			int n = sprintf( ptr, "{%d}", bdn->a_level );
+			char buf[SLAP_TEXT_BUFLEN];
+			int n = snprintf( buf, sizeof(buf), "{%d}", bdn->a_level );
 			if ( n > 0 ) {
-				ptr += n;
+				ptr = acl_safe_strncopy( ptr, buf, n );
 			} /* else ? */
 		}
 		if ( bdn->a_expand ) {
-			ptr = lutil_strcopy( ptr, ",expand" );
+			ptr = acl_safe_strcopy( ptr, ",expand" );
 		}
-		*ptr++ = '=';
-		*ptr++ = '"';
-		ptr = lutil_strcopy( ptr, bdn->a_pat.bv_val );
-		*ptr++ = '"';
+		ptr = acl_safe_strcopy( ptr, "=\"" );
+		ptr = acl_safe_strbvcopy( ptr, &bdn->a_pat );
+		ptr = acl_safe_strcopy( ptr, "\"" );
 	}
 	return ptr;
 }
@@ -2585,88 +2570,83 @@ access2text( Access *b, char *ptr )
 {
 	char maskbuf[ACCESSMASK_MAXLEN];
 
-	ptr = lutil_strcopy( ptr, "\tby" );
+	ptr = acl_safe_strcopy( ptr, "\tby" );
 
 	if ( !BER_BVISEMPTY( &b->a_dn_pat ) ) {
 		ptr = dnaccess2text( &b->a_dn, ptr, 0 );
 	}
 	if ( b->a_dn_at ) {
-		ptr = lutil_strcopy( ptr, " dnattr=" );
-		ptr = lutil_strcopy( ptr, b->a_dn_at->ad_cname.bv_val );
+		ptr = acl_safe_strcopy( ptr, " dnattr=" );
+		ptr = acl_safe_strbvcopy( ptr, &b->a_dn_at->ad_cname );
 	}
 
 	if ( !BER_BVISEMPTY( &b->a_realdn_pat ) ) {
 		ptr = dnaccess2text( &b->a_realdn, ptr, 1 );
 	}
 	if ( b->a_realdn_at ) {
-		ptr = lutil_strcopy( ptr, " realdnattr=" );
-		ptr = lutil_strcopy( ptr, b->a_realdn_at->ad_cname.bv_val );
+		ptr = acl_safe_strcopy( ptr, " realdnattr=" );
+		ptr = acl_safe_strbvcopy( ptr, &b->a_realdn_at->ad_cname );
 	}
 
 	if ( !BER_BVISEMPTY( &b->a_group_pat ) ) {
-		ptr = lutil_strcopy( ptr, " group/" );
-		ptr = lutil_strcopy( ptr, b->a_group_oc ?
+		ptr = acl_safe_strcopy( ptr, " group/" );
+		ptr = acl_safe_strcopy( ptr, b->a_group_oc ?
 			b->a_group_oc->soc_cname.bv_val : SLAPD_GROUP_CLASS );
-		*ptr++ = '/';
-		ptr = lutil_strcopy( ptr, b->a_group_at ?
+		ptr = acl_safe_strcopy( ptr, "/" );
+		ptr = acl_safe_strcopy( ptr, b->a_group_at ?
 			b->a_group_at->ad_cname.bv_val : SLAPD_GROUP_ATTR );
-		*ptr++ = '.';
-		ptr = lutil_strcopy( ptr, style_strings[b->a_group_style] );
-		*ptr++ = '=';
-		*ptr++ = '"';
-		ptr = lutil_strcopy( ptr, b->a_group_pat.bv_val );
-		*ptr++ = '"';
+		ptr = acl_safe_strcopy( ptr, "." );
+		ptr = acl_safe_strcopy( ptr, style_strings[b->a_group_style] );
+		ptr = acl_safe_strcopy( ptr, "=\"" );
+		ptr = acl_safe_strbvcopy( ptr, &b->a_group_pat );
+		ptr = acl_safe_strcopy( ptr, "\"" );
 	}
 
 	if ( !BER_BVISEMPTY( &b->a_peername_pat ) ) {
-		ptr = lutil_strcopy( ptr, " peername" );
-		*ptr++ = '.';
-		ptr = lutil_strcopy( ptr, style_strings[b->a_peername_style] );
-		*ptr++ = '=';
-		*ptr++ = '"';
-		ptr = lutil_strcopy( ptr, b->a_peername_pat.bv_val );
-		*ptr++ = '"';
+		ptr = acl_safe_strcopy( ptr, " peername" );
+		ptr = acl_safe_strcopy( ptr, "." );
+		ptr = acl_safe_strcopy( ptr, style_strings[b->a_peername_style] );
+		ptr = acl_safe_strcopy( ptr, "=\"" );
+		ptr = acl_safe_strbvcopy( ptr, &b->a_peername_pat );
+		ptr = acl_safe_strcopy( ptr, "\"" );
 	}
 
 	if ( !BER_BVISEMPTY( &b->a_sockname_pat ) ) {
-		ptr = lutil_strcopy( ptr, " sockname" );
-		*ptr++ = '.';
-		ptr = lutil_strcopy( ptr, style_strings[b->a_sockname_style] );
-		*ptr++ = '=';
-		*ptr++ = '"';
-		ptr = lutil_strcopy( ptr, b->a_sockname_pat.bv_val );
-		*ptr++ = '"';
+		ptr = acl_safe_strcopy( ptr, " sockname" );
+		ptr = acl_safe_strcopy( ptr, "." );
+		ptr = acl_safe_strcopy( ptr, style_strings[b->a_sockname_style] );
+		ptr = acl_safe_strcopy( ptr, "=\"" );
+		ptr = acl_safe_strbvcopy( ptr, &b->a_sockname_pat );
+		ptr = acl_safe_strcopy( ptr, "\"" );
 	}
 
 	if ( !BER_BVISEMPTY( &b->a_domain_pat ) ) {
-		ptr = lutil_strcopy( ptr, " domain" );
-		*ptr++ = '.';
-		ptr = lutil_strcopy( ptr, style_strings[b->a_domain_style] );
+		ptr = acl_safe_strcopy( ptr, " domain" );
+		ptr = acl_safe_strcopy( ptr, "." );
+		ptr = acl_safe_strcopy( ptr, style_strings[b->a_domain_style] );
 		if ( b->a_domain_expand ) {
-			ptr = lutil_strcopy( ptr, ",expand" );
+			ptr = acl_safe_strcopy( ptr, ",expand" );
 		}
-		*ptr++ = '=';
-		ptr = lutil_strcopy( ptr, b->a_domain_pat.bv_val );
+		ptr = acl_safe_strcopy( ptr, "=" );
+		ptr = acl_safe_strbvcopy( ptr, &b->a_domain_pat );
 	}
 
 	if ( !BER_BVISEMPTY( &b->a_sockurl_pat ) ) {
-		ptr = lutil_strcopy( ptr, " sockurl" );
-		*ptr++ = '.';
-		ptr = lutil_strcopy( ptr, style_strings[b->a_sockurl_style] );
-		*ptr++ = '=';
-		*ptr++ = '"';
-		ptr = lutil_strcopy( ptr, b->a_sockurl_pat.bv_val );
-		*ptr++ = '"';
+		ptr = acl_safe_strcopy( ptr, " sockurl" );
+		ptr = acl_safe_strcopy( ptr, "." );
+		ptr = acl_safe_strcopy( ptr, style_strings[b->a_sockurl_style] );
+		ptr = acl_safe_strcopy( ptr, "=\"" );
+		ptr = acl_safe_strbvcopy( ptr, &b->a_sockurl_pat );
+		ptr = acl_safe_strcopy( ptr, "\"" );
 	}
 
 	if ( !BER_BVISEMPTY( &b->a_set_pat ) ) {
-		ptr = lutil_strcopy( ptr, " set" );
-		*ptr++ = '.';
-		ptr = lutil_strcopy( ptr, style_strings[b->a_set_style] );
-		*ptr++ = '=';
-		*ptr++ = '"';
-		ptr = lutil_strcopy( ptr, b->a_set_pat.bv_val );
-		*ptr++ = '"';
+		ptr = acl_safe_strcopy( ptr, " set" );
+		ptr = acl_safe_strcopy( ptr, "." );
+		ptr = acl_safe_strcopy( ptr, style_strings[b->a_set_style] );
+		ptr = acl_safe_strcopy( ptr, "=\"" );
+		ptr = acl_safe_strbvcopy( ptr, &b->a_set_pat );
+		ptr = acl_safe_strcopy( ptr, "\"" );
 	}
 
 #ifdef SLAP_DYNACL
@@ -2678,7 +2658,7 @@ access2text( Access *b, char *ptr )
 				struct berval bv = BER_BVNULL;
 				(void)( *da->da_unparse )( da->da_private, &bv );
 				assert( !BER_BVISNULL( &bv ) );
-				ptr = lutil_strcopy( ptr, bv.bv_val );
+				ptr = acl_safe_strbvcopy( ptr, &bv );
 				ch_free( bv.bv_val );
 			}
 		}
@@ -2687,43 +2667,51 @@ access2text( Access *b, char *ptr )
 
 	/* Security Strength Factors */
 	if ( b->a_authz.sai_ssf ) {
-		ptr += sprintf( ptr, " ssf=%u", 
+		char buf[SLAP_TEXT_BUFLEN];
+		int n = snprintf( buf, sizeof(buf), " ssf=%u", 
 			b->a_authz.sai_ssf );
+		ptr = acl_safe_strncopy( ptr, buf, n );
 	}
 	if ( b->a_authz.sai_transport_ssf ) {
-		ptr += sprintf( ptr, " transport_ssf=%u",
+		char buf[SLAP_TEXT_BUFLEN];
+		int n = snprintf( buf, sizeof(buf), " transport_ssf=%u",
 			b->a_authz.sai_transport_ssf );
+		ptr = acl_safe_strncopy( ptr, buf, n );
 	}
 	if ( b->a_authz.sai_tls_ssf ) {
-		ptr += sprintf( ptr, " tls_ssf=%u",
+		char buf[SLAP_TEXT_BUFLEN];
+		int n = snprintf( buf, sizeof(buf), " tls_ssf=%u",
 			b->a_authz.sai_tls_ssf );
+		ptr = acl_safe_strncopy( ptr, buf, n );
 	}
 	if ( b->a_authz.sai_sasl_ssf ) {
-		ptr += sprintf( ptr, " sasl_ssf=%u",
+		char buf[SLAP_TEXT_BUFLEN];
+		int n = snprintf( buf, sizeof(buf), " sasl_ssf=%u",
 			b->a_authz.sai_sasl_ssf );
+		ptr = acl_safe_strncopy( ptr, buf, n );
 	}
 
-	*ptr++ = ' ';
+	ptr = acl_safe_strcopy( ptr, " " );
 	if ( b->a_dn_self ) {
-		ptr = lutil_strcopy( ptr, "self" );
+		ptr = acl_safe_strcopy( ptr, "self" );
 	} else if ( b->a_realdn_self ) {
-		ptr = lutil_strcopy( ptr, "realself" );
+		ptr = acl_safe_strcopy( ptr, "realself" );
 	}
-	ptr = lutil_strcopy( ptr, accessmask2str( b->a_access_mask, maskbuf, 0 ));
+	ptr = acl_safe_strcopy( ptr, accessmask2str( b->a_access_mask, maskbuf, 0 ));
 	if ( !maskbuf[0] ) ptr--;
 
 	if( b->a_type == ACL_BREAK ) {
-		ptr = lutil_strcopy( ptr, " break" );
+		ptr = acl_safe_strcopy( ptr, " break" );
 
 	} else if( b->a_type == ACL_CONTINUE ) {
-		ptr = lutil_strcopy( ptr, " continue" );
+		ptr = acl_safe_strcopy( ptr, " continue" );
 
 	} else if( b->a_type != ACL_STOP ) {
-		ptr = lutil_strcopy( ptr, " unknown-control" );
+		ptr = acl_safe_strcopy( ptr, " unknown-control" );
 	} else {
-		if ( !maskbuf[0] ) ptr = lutil_strcopy( ptr, " stop" );
+		if ( !maskbuf[0] ) ptr = acl_safe_strcopy( ptr, " stop" );
 	}
-	*ptr++ = '\n';
+	ptr = acl_safe_strcopy( ptr, "\n" );
 
 	return ptr;
 }
@@ -2735,35 +2723,37 @@ acl_unparse( AccessControl *a, struct berval *bv )
 	char	*ptr;
 	int	to = 0;
 
-	bv->bv_val = aclbuf;
+	if ( BER_BVISNULL( &aclbuf ) ) {
+		aclbuf.bv_val = ch_malloc( ACLBUF_CHUNKSIZE );
+		aclbuf.bv_len = ACLBUF_CHUNKSIZE;
+	}
+
 	bv->bv_len = 0;
 
-	ptr = bv->bv_val;
+	ptr = aclbuf.bv_val;
 
-	ptr = lutil_strcopy( ptr, "to" );
+	ptr = acl_safe_strcopy( ptr, "to" );
 	if ( !BER_BVISNULL( &a->acl_dn_pat ) ) {
 		to++;
-		ptr = lutil_strcopy( ptr, " dn." );
+		ptr = acl_safe_strcopy( ptr, " dn." );
 		if ( a->acl_dn_style == ACL_STYLE_BASE )
-			ptr = lutil_strcopy( ptr, style_base );
+			ptr = acl_safe_strcopy( ptr, style_base );
 		else
-			ptr = lutil_strcopy( ptr, style_strings[a->acl_dn_style] );
-		*ptr++ = '=';
-		*ptr++ = '"';
-		ptr = lutil_strcopy( ptr, a->acl_dn_pat.bv_val );
-		ptr = lutil_strcopy( ptr, "\"\n" );
+			ptr = acl_safe_strcopy( ptr, style_strings[a->acl_dn_style] );
+		ptr = acl_safe_strcopy( ptr, "=\"" );
+		ptr = acl_safe_strbvcopy( ptr, &a->acl_dn_pat );
+		ptr = acl_safe_strcopy( ptr, "\"\n" );
 	}
 
 	if ( a->acl_filter != NULL ) {
-		struct berval	bv = BER_BVNULL;
+		struct berval	fbv = BER_BVNULL;
 
 		to++;
-		filter2bv( a->acl_filter, &bv );
-		ptr = lutil_strcopy( ptr, " filter=\"" );
-		ptr = lutil_strcopy( ptr, bv.bv_val );
-		*ptr++ = '"';
-		*ptr++ = '\n';
-		ch_free( bv.bv_val );
+		filter2bv( a->acl_filter, &fbv );
+		ptr = acl_safe_strcopy( ptr, " filter=\"" );
+		ptr = acl_safe_strbvcopy( ptr, &fbv );
+		ptr = acl_safe_strcopy( ptr, "\"\n" );
+		ch_free( fbv.bv_val );
 	}
 
 	if ( a->acl_attrs != NULL ) {
@@ -2771,45 +2761,44 @@ acl_unparse( AccessControl *a, struct berval *bv )
 		AttributeName *an;
 		to++;
 
-		ptr = lutil_strcopy( ptr, " attrs=" );
+		ptr = acl_safe_strcopy( ptr, " attrs=" );
 		for ( an = a->acl_attrs; an && !BER_BVISNULL( &an->an_name ); an++ ) {
-			if ( ! first ) *ptr++ = ',';
+			if ( ! first ) ptr = acl_safe_strcopy( ptr, ",");
 			if (an->an_oc) {
-				*ptr++ = ( an->an_flags & SLAP_AN_OCEXCLUDE ) ? '!' : '@';
-				ptr = lutil_strcopy( ptr, an->an_oc->soc_cname.bv_val );
+				ptr = acl_safe_strcopy( ptr, ( an->an_flags & SLAP_AN_OCEXCLUDE ) ? "!" : "@" );
+				ptr = acl_safe_strbvcopy( ptr, &an->an_oc->soc_cname );
 
 			} else {
-				ptr = lutil_strcopy( ptr, an->an_name.bv_val );
+				ptr = acl_safe_strbvcopy( ptr, &an->an_name );
 			}
 			first = 0;
 		}
-		*ptr++ = '\n';
+		ptr = acl_safe_strcopy( ptr, "\n" );
 	}
 
-	if ( !BER_BVISEMPTY( &a->acl_attrval ) ) {
+	if ( !BER_BVISNULL( &a->acl_attrval ) ) {
 		to++;
-		ptr = lutil_strcopy( ptr, " val." );
+		ptr = acl_safe_strcopy( ptr, " val." );
 		if ( a->acl_attrval_style == ACL_STYLE_BASE &&
 			a->acl_attrs[0].an_desc->ad_type->sat_syntax ==
 				slap_schema.si_syn_distinguishedName )
-			ptr = lutil_strcopy( ptr, style_base );
+			ptr = acl_safe_strcopy( ptr, style_base );
 		else
-			ptr = lutil_strcopy( ptr, style_strings[a->acl_attrval_style] );
-		*ptr++ = '=';
-		*ptr++ = '"';
-		ptr = lutil_strcopy( ptr, a->acl_attrval.bv_val );
-		*ptr++ = '"';
-		*ptr++ = '\n';
+			ptr = acl_safe_strcopy( ptr, style_strings[a->acl_attrval_style] );
+		ptr = acl_safe_strcopy( ptr, "=\"" );
+		ptr = acl_safe_strbvcopy( ptr, &a->acl_attrval );
+		ptr = acl_safe_strcopy( ptr, "\"\n" );
 	}
 
-	if( !to ) {
-		ptr = lutil_strcopy( ptr, " *\n" );
+	if ( !to ) {
+		ptr = acl_safe_strcopy( ptr, " *\n" );
 	}
 
 	for ( b = a->acl_access; b != NULL; b = b->a_next ) {
 		ptr = access2text( b, ptr );
 	}
 	*ptr = '\0';
+	bv->bv_val = aclbuf.bv_val;
 	bv->bv_len = ptr - bv->bv_val;
 }
 
