@@ -1,5 +1,5 @@
 /* memberof.c - back-reference for group membership */
-/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/memberof.c,v 1.2.2.21 2010/04/15 20:11:24 quanah Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
  * Copyright 2005-2007 Pierangelo Masarati <ando@sys-net.it>
@@ -28,7 +28,7 @@
 #include "ac/socket.h"
 
 #include "slap.h"
-#include "config.h"
+#include "slap-config.h"
 #include "lutil.h"
 
 /*
@@ -130,6 +130,11 @@
 
 #define	SLAPD_MEMBEROF_ATTR	"memberOf"
 
+static AttributeDescription	*ad_member;
+static AttributeDescription	*ad_memberOf;
+
+static ObjectClass			*oc_group;
+
 static slap_overinst		memberof;
 
 typedef struct memberof_t {
@@ -190,7 +195,16 @@ typedef struct memberof_cbinfo_t {
 	BerVarray memberof;
 	memberof_is_t what;
 } memberof_cbinfo_t;
-	
+
+static void
+memberof_set_backend( Operation *op_target, Operation *op, slap_overinst *on )
+{
+    BackendInfo *bi = op->o_bd->bd_info;
+
+    if ( bi->bi_type == memberof.on_bi.bi_type )
+        op_target->o_bd->bd_info = (BackendInfo *)on->on_info;
+}
+
 static int
 memberof_isGroupOrMember_cb( Operation *op, SlapReply *rs )
 {
@@ -243,7 +257,6 @@ memberof_isGroupOrMember( Operation *op, memberof_cbinfo_t *mci )
 	memberof_t		*mo = (memberof_t *)on->on_bi.bi_private;
 
 	Operation		op2 = *op;
-	SlapReply		rs2 = { REP_RESULT };
 	slap_callback		cb = { 0 };
 	BackendInfo	*bi = op->o_bd->bd_info;
 	AttributeName		an[ 2 ];
@@ -276,6 +289,8 @@ memberof_isGroupOrMember( Operation *op, memberof_cbinfo_t *mci )
 	op2.ors_tlimit = SLAP_NO_LIMIT;
 
 	if ( mci->what & MEMBEROF_IS_GROUP ) {
+		SlapReply	rs2 = { REP_RESULT };
+
 		mc.ad = mo->mo_ad_member;
 		mc.foundit = 0;
 		mc.vals = NULL;
@@ -283,8 +298,9 @@ memberof_isGroupOrMember( Operation *op, memberof_cbinfo_t *mci )
 		an[ 0 ].an_name = an[ 0 ].an_desc->ad_cname;
 		op2.ors_filterstr = mo->mo_groupFilterstr;
 		op2.ors_filter = &mo->mo_groupFilter;
+		op2.o_do_not_cache = 1;	/* internal search, don't log */
 
-		op2.o_bd->bd_info = (BackendInfo *)on->on_info;
+		memberof_set_backend( &op2, op, on );
 		(void)op->o_bd->be_search( &op2, &rs2 );
 		op2.o_bd->bd_info = bi;
 
@@ -296,6 +312,8 @@ memberof_isGroupOrMember( Operation *op, memberof_cbinfo_t *mci )
 	}
 
 	if ( mci->what & MEMBEROF_IS_MEMBER ) {
+		SlapReply	rs2 = { REP_RESULT };
+
 		mc.ad = mo->mo_ad_memberof;
 		mc.foundit = 0;
 		mc.vals = NULL;
@@ -303,8 +321,9 @@ memberof_isGroupOrMember( Operation *op, memberof_cbinfo_t *mci )
 		an[ 0 ].an_name = an[ 0 ].an_desc->ad_cname;
 		op2.ors_filterstr = mo->mo_memberFilterstr;
 		op2.ors_filter = &mo->mo_memberFilter;
+		op2.o_do_not_cache = 1;	/* internal search, don't log */
 
-		op2.o_bd->bd_info = (BackendInfo *)on->on_info;
+		memberof_set_backend( &op2, op, on );
 		(void)op->o_bd->be_search( &op2, &rs2 );
 		op2.o_bd->bd_info = bi;
 
@@ -323,10 +342,9 @@ memberof_isGroupOrMember( Operation *op, memberof_cbinfo_t *mci )
 /*
  * response callback that adds memberof values when a group is modified.
  */
-static int
+static void
 memberof_value_modify(
 	Operation		*op,
-	SlapReply		*rs,
 	struct berval		*ndn,
 	AttributeDescription	*ad,
 	struct berval		*old_dn,
@@ -339,12 +357,19 @@ memberof_value_modify(
 	memberof_t	*mo = (memberof_t *)on->on_bi.bi_private;
 
 	Operation	op2 = *op;
+	unsigned long opid = op->o_opid;
 	SlapReply	rs2 = { REP_RESULT };
 	slap_callback	cb = { NULL, slap_null_cb, NULL, NULL };
 	Modifications	mod[ 2 ] = { { { 0 } } }, *ml;
 	struct berval	values[ 4 ], nvalues[ 4 ];
 	int		mcnt = 0;
 
+	if ( old_ndn != NULL && new_ndn != NULL &&
+		ber_bvcmp( old_ndn, new_ndn ) == 0 ) {
+	    /* DNs compare equal, it's a noop */
+	    return;
+	}
+
 	op2.o_tag = LDAP_REQ_MODIFY;
 
 	op2.o_req_dn = *ndn;
@@ -355,6 +380,11 @@ memberof_value_modify(
 	op2.o_ndn = op->o_bd->be_rootndn;
 	op2.orm_modlist = NULL;
 
+	/* Internal ops, never replicate these */
+	op2.o_opid = 0;		/* shared with op, saved above */
+	op2.orm_no_opattrs = 1;
+	op2.o_dont_replicate = 1;
+
 	if ( !BER_BVISNULL( &mo->mo_ndn ) ) {
 		ml = &mod[ mcnt ];
 		ml->sml_numvals = 1;
@@ -385,9 +415,11 @@ memberof_value_modify(
 	ml->sml_flags = SLAP_MOD_INTERNAL;
 	ml->sml_next = op2.orm_modlist;
 	op2.orm_modlist = ml;
-	op2.orm_no_opattrs = 0;
 
 	if ( new_ndn != NULL ) {
+		BackendInfo *bi = op2.o_bd->bd_info;
+		OpExtra	oex;
+
 		assert( !BER_BVISNULL( new_dn ) );
 		assert( !BER_BVISNULL( new_ndn ) );
 
@@ -397,15 +429,17 @@ memberof_value_modify(
 		ml->sml_values[ 0 ] = *new_dn;
 		ml->sml_nvalues[ 0 ] = *new_ndn;
 
+		oex.oe_key = (void *)&memberof;
+		LDAP_SLIST_INSERT_HEAD(&op2.o_extra, &oex, oe_next);
+		memberof_set_backend( &op2, op, on );
 		(void)op->o_bd->be_modify( &op2, &rs2 );
+		op2.o_bd->bd_info = bi;
+		LDAP_SLIST_REMOVE(&op2.o_extra, &oex, OpExtra, oe_next);
 		if ( rs2.sr_err != LDAP_SUCCESS ) {
-			char buf[ SLAP_TEXT_BUFLEN ];
-			snprintf( buf, sizeof( buf ),
-				"memberof_value_modify %s=\"%s\" failed err=%d text=%s",
-				ad->ad_cname.bv_val, new_dn->bv_val, rs2.sr_err,
-				rs2.sr_text ? rs2.sr_text : "" );
-			Debug( LDAP_DEBUG_ANY, "%s: %s\n",
-				op->o_log_prefix, buf, 0 );
+			Debug(LDAP_DEBUG_ANY,
+			      "%s: memberof_value_modify DN=\"%s\" add %s=\"%s\" failed err=%d\n",
+			      op->o_log_prefix, op2.o_req_dn.bv_val,
+			      ad->ad_cname.bv_val, new_dn->bv_val, rs2.sr_err );
 		}
 
 		assert( op2.orm_modlist == &mod[ mcnt ] );
@@ -423,6 +457,9 @@ memberof_value_modify(
 	}
 
 	if ( old_ndn != NULL ) {
+		BackendInfo *bi = op2.o_bd->bd_info;
+		OpExtra	oex;
+
 		assert( !BER_BVISNULL( old_dn ) );
 		assert( !BER_BVISNULL( old_ndn ) );
 
@@ -432,15 +469,17 @@ memberof_value_modify(
 		ml->sml_values[ 0 ] = *old_dn;
 		ml->sml_nvalues[ 0 ] = *old_ndn;
 
+		oex.oe_key = (void *)&memberof;
+		LDAP_SLIST_INSERT_HEAD(&op2.o_extra, &oex, oe_next);
+		memberof_set_backend( &op2, op, on );
 		(void)op->o_bd->be_modify( &op2, &rs2 );
+		op2.o_bd->bd_info = bi;
+		LDAP_SLIST_REMOVE(&op2.o_extra, &oex, OpExtra, oe_next);
 		if ( rs2.sr_err != LDAP_SUCCESS ) {
-			char buf[ SLAP_TEXT_BUFLEN ];
-			snprintf( buf, sizeof( buf ),
-				"memberof_value_modify %s=\"%s\" failed err=%d text=%s",
-				ad->ad_cname.bv_val, old_dn->bv_val, rs2.sr_err,
-				rs2.sr_text ? rs2.sr_text : "" );
-			Debug( LDAP_DEBUG_ANY, "%s: %s\n",
-				op->o_log_prefix, buf, 0 );
+			Debug(LDAP_DEBUG_ANY,
+			      "%s: memberof_value_modify DN=\"%s\" delete %s=\"%s\" failed err=%d\n",
+			      op->o_log_prefix, op2.o_req_dn.bv_val,
+			      ad->ad_cname.bv_val, old_dn->bv_val, rs2.sr_err );
 		}
 
 		assert( op2.orm_modlist == &mod[ mcnt ] );
@@ -453,13 +492,13 @@ memberof_value_modify(
 			slap_mods_free( ml, 1 );
 		}
 	}
+	/* restore original opid */
+	op->o_opid = opid;
 
 	/* FIXME: if old_group_ndn doesn't exist, both delete __and__
 	 * add will fail; better split in two operations, although
 	 * not optimal in terms of performance.  At least it would
 	 * move towards self-repairing capabilities. */
-
-	return rs2.sr_err;
 }
 
 static int
@@ -494,13 +533,19 @@ memberof_op_add( Operation *op, SlapReply *rs )
 	struct berval	save_dn, save_ndn;
 	slap_callback *sc;
 	memberof_cbinfo_t *mci;
+	OpExtra		*oex;
+
+	LDAP_SLIST_FOREACH( oex, &op->o_extra, oe_next ) {
+		if ( oex->oe_key == (void *)&memberof )
+			return SLAP_CB_CONTINUE;
+	}
 
 	if ( op->ora_e->e_attrs == NULL ) {
 		/* FIXME: global overlay; need to deal with */
 		Debug( LDAP_DEBUG_ANY, "%s: memberof_op_add(\"%s\"): "
 			"consistency checks not implemented when overlay "
 			"is instantiated as global.\n",
-			op->o_log_prefix, op->o_req_dn.bv_val, 0 );
+			op->o_log_prefix, op->o_req_dn.bv_val );
 		return SLAP_CB_CONTINUE;
 	}
 
@@ -539,6 +584,10 @@ memberof_op_add( Operation *op, SlapReply *rs )
 			for ( i = 0; !BER_BVISNULL( &a->a_nvals[ i ] ); i++ ) {
 				Entry		*e = NULL;
 
+				/* ITS#6670 Ignore member pointing to this entry */
+				if ( dn_match( &a->a_nvals[i], &save_ndn ))
+					continue;
+
 				rc = be_entry_get_rw( op, &a->a_nvals[ i ],
 						NULL, NULL, 0, &e );
 				if ( rc == LDAP_SUCCESS ) {
@@ -569,6 +618,7 @@ memberof_op_add( Operation *op, SlapReply *rs )
 						ber_memfree( a->a_nvals[ i ].bv_val );
 						BER_BVZERO( &a->a_nvals[ i ] );
 					}
+					a->a_numvals--;
 					if ( j - i == 1 ) {
 						break;
 					}
@@ -580,7 +630,6 @@ memberof_op_add( Operation *op, SlapReply *rs )
 							sizeof( struct berval ) * ( j - i ) );
 					}
 					i--;
-					a->a_numvals--;
 				}
 			}
 
@@ -617,6 +666,10 @@ memberof_op_add( Operation *op, SlapReply *rs )
 				send_ldap_result( op, rs );
 				goto done;
 			}
+			/* ITS#6670 Ignore member pointing to this entry */
+			if ( dn_match( &a->a_nvals[i], &save_ndn ))
+				continue;
+
 			rc = be_entry_get_rw( op, &a->a_nvals[ i ],
 					NULL, NULL, 0, &e );
 			op->o_bd->bd_info = (BackendInfo *)on;
@@ -691,6 +744,7 @@ memberof_op_add( Operation *op, SlapReply *rs )
 	sc->sc_private = sc+1;
 	sc->sc_response = memberof_res_add;
 	sc->sc_cleanup = memberof_cleanup;
+	sc->sc_writewait = 0;
 	mci = sc->sc_private;
 	mci->on = on;
 	mci->member = NULL;
@@ -714,12 +768,18 @@ memberof_op_delete( Operation *op, SlapReply *rs )
 
 	slap_callback *sc;
 	memberof_cbinfo_t *mci;
+	OpExtra		*oex;
 
+	LDAP_SLIST_FOREACH( oex, &op->o_extra, oe_next ) {
+		if ( oex->oe_key == (void *)&memberof )
+			return SLAP_CB_CONTINUE;
+	}
 
 	sc = op->o_tmpalloc( sizeof(slap_callback)+sizeof(*mci), op->o_tmpmemctx );
 	sc->sc_private = sc+1;
 	sc->sc_response = memberof_res_delete;
 	sc->sc_cleanup = memberof_cleanup;
+	sc->sc_writewait = 0;
 	mci = sc->sc_private;
 	mci->on = on;
 	mci->member = NULL;
@@ -748,6 +808,12 @@ memberof_op_modify( Operation *op, SlapReply *rs )
 	struct berval	save_dn, save_ndn;
 	slap_callback *sc;
 	memberof_cbinfo_t *mci, mcis;
+	OpExtra		*oex;
+
+	LDAP_SLIST_FOREACH( oex, &op->o_extra, oe_next ) {
+		if ( oex->oe_key == (void *)&memberof )
+			return SLAP_CB_CONTINUE;
+	}
 
 	if ( MEMBEROF_REVERSE( mo ) ) {
 		for ( mlp = &op->orm_modlist; *mlp; mlp = &(*mlp)->sml_next ) {
@@ -775,6 +841,7 @@ memberof_op_modify( Operation *op, SlapReply *rs )
 				switch ( ml->sml_op ) {
 				case LDAP_MOD_DELETE:
 				case LDAP_MOD_REPLACE:
+				case SLAP_MOD_SOFTDEL: /* ITS#7487: can be used by syncrepl (in mirror mode?) */
 					save_member = 1;
 					break;
 				}
@@ -802,6 +869,7 @@ memberof_op_modify( Operation *op, SlapReply *rs )
 		
 				switch ( ml->sml_op ) {
 				case LDAP_MOD_DELETE:
+				case SLAP_MOD_SOFTDEL: /* ITS#7487: can be used by syncrepl (in mirror mode?) */
 					/* we don't care about cancellations: if the value
 					 * exists, fine; if it doesn't, we let the underlying
 					 * database fail as appropriate; */
@@ -816,14 +884,19 @@ memberof_op_modify( Operation *op, SlapReply *rs )
  					}
  
 				case LDAP_MOD_ADD:
+				case SLAP_MOD_SOFTADD: /* ITS#7487 */
+				case SLAP_MOD_ADD_IF_NOT_PRESENT: /* ITS#7487 */
 					/* NOTE: right now, the attributeType we use
 					 * for member must have a normalized value */
 					assert( ml->sml_nvalues != NULL );
 		
 					for ( i = 0; !BER_BVISNULL( &ml->sml_nvalues[ i ] ); i++ ) {
-						int		rc;
 						Entry		*e;
 		
+						/* ITS#6670 Ignore member pointing to this entry */
+						if ( dn_match( &ml->sml_nvalues[i], &save_ndn ))
+							continue;
+
 						if ( be_entry_get_rw( op, &ml->sml_nvalues[ i ],
 								NULL, NULL, 0, &e ) == LDAP_SUCCESS )
 						{
@@ -900,6 +973,7 @@ memberof_op_modify( Operation *op, SlapReply *rs )
 
 		switch ( ml->sml_op ) {
 		case LDAP_MOD_DELETE:
+		case SLAP_MOD_SOFTDEL: /* ITS#7487: can be used by syncrepl (in mirror mode?) */
 			if ( ml->sml_nvalues != NULL ) {
 				AccessControlState	acl_state = ACL_STATE_INIT;
 
@@ -1010,12 +1084,15 @@ memberof_op_modify( Operation *op, SlapReply *rs )
 				goto done2;
 			}
 
-			if ( ml->sml_op == LDAP_MOD_DELETE || !ml->sml_values ) {
+			if ( ml->sml_op == LDAP_MOD_DELETE || ml->sml_op == SLAP_MOD_SOFTDEL || !ml->sml_values ) {
 				break;
 			}
 			/* fall thru */
 
-		case LDAP_MOD_ADD: {
+		case LDAP_MOD_ADD:
+		case SLAP_MOD_SOFTADD: /* ITS#7487 */
+		case SLAP_MOD_ADD_IF_NOT_PRESENT: /* ITS#7487 */
+			{
 			AccessControlState	acl_state = ACL_STATE_INIT;
 
 			for ( i = 0; !BER_BVISNULL( &ml->sml_nvalues[ i ] ); i++ ) {
@@ -1035,6 +1112,10 @@ memberof_op_modify( Operation *op, SlapReply *rs )
 					goto done2;
 				}
 
+				/* ITS#6670 Ignore member pointing to this entry */
+				if ( dn_match( &ml->sml_nvalues[i], &save_ndn ))
+					continue;
+
 				rc = be_entry_get_rw( op, &ml->sml_nvalues[ i ],
 						NULL, NULL, 0, &e );
 				op->o_bd->bd_info = (BackendInfo *)on;
@@ -1117,6 +1198,7 @@ done2:;
 	sc->sc_private = sc+1;
 	sc->sc_response = memberof_res_modify;
 	sc->sc_cleanup = memberof_cleanup;
+	sc->sc_writewait = 0;
 	mci = sc->sc_private;
 	mci->on = on;
 	mci->member = NULL;
@@ -1151,11 +1233,18 @@ memberof_op_modrdn( Operation *op, SlapReply *rs )
 	slap_overinst	*on = (slap_overinst *)op->o_bd->bd_info;
 	slap_callback *sc;
 	memberof_cbinfo_t *mci;
+	OpExtra		*oex;
+
+	LDAP_SLIST_FOREACH( oex, &op->o_extra, oe_next ) {
+		if ( oex->oe_key == (void *)&memberof )
+			return SLAP_CB_CONTINUE;
+	}
 
 	sc = op->o_tmpalloc( sizeof(slap_callback)+sizeof(*mci), op->o_tmpmemctx );
 	sc->sc_private = sc+1;
 	sc->sc_response = memberof_res_modrdn;
 	sc->sc_cleanup = memberof_cleanup;
+	sc->sc_writewait = 0;
 	mci = sc->sc_private;
 	mci->on = on;
 	mci->member = NULL;
@@ -1188,17 +1277,19 @@ memberof_res_add( Operation *op, SlapReply *rs )
 
 		ma = attr_find( op->ora_e->e_attrs, mo->mo_ad_memberof );
 		if ( ma != NULL ) {
-			char relax = op->o_relax;
-
 			/* relax is required to allow to add
 			 * a non-existing member */
 			op->o_relax = SLAP_CONTROL_CRITICAL;
 
 			for ( i = 0; !BER_BVISNULL( &ma->a_nvals[ i ] ); i++ ) {
 		
+				/* ITS#6670 Ignore member pointing to this entry */
+				if ( dn_match( &ma->a_nvals[i], &op->o_req_ndn ))
+					continue;
+
 				/* the modification is attempted
 				 * with the original identity */
-				(void)memberof_value_modify( op, rs,
+				memberof_value_modify( op,
 					&ma->a_nvals[ i ], mo->mo_ad_member,
 					NULL, NULL, &op->o_req_dn, &op->o_req_ndn );
 			}
@@ -1213,7 +1304,11 @@ memberof_res_add( Operation *op, SlapReply *rs )
 				a = attrs_find( a->a_next, mo->mo_ad_member ) )
 		{
 			for ( i = 0; !BER_BVISNULL( &a->a_nvals[ i ] ); i++ ) {
-				(void)memberof_value_modify( op, rs,
+				/* ITS#6670 Ignore member pointing to this entry */
+				if ( dn_match( &a->a_nvals[i], &op->o_req_ndn ))
+					continue;
+
+				memberof_value_modify( op,
 						&a->a_nvals[ i ],
 						mo->mo_ad_memberof,
 						NULL, NULL,
@@ -1246,7 +1341,7 @@ memberof_res_delete( Operation *op, SlapReply *rs )
 	vals = mci->member;
 	if ( vals != NULL ) {
 		for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-			(void)memberof_value_modify( op, rs,
+			memberof_value_modify( op,
 					&vals[ i ], mo->mo_ad_memberof,
 					&op->o_req_dn, &op->o_req_ndn,
 					NULL, NULL );
@@ -1257,7 +1352,7 @@ memberof_res_delete( Operation *op, SlapReply *rs )
 		vals = mci->memberof;
 		if ( vals != NULL ) {
 			for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-				(void)memberof_value_modify( op, rs,
+				memberof_value_modify( op,
 						&vals[ i ], mo->mo_ad_member,
 						&op->o_req_dn, &op->o_req_ndn,
 						NULL, NULL );
@@ -1301,9 +1396,10 @@ memberof_res_modify( Operation *op, SlapReply *rs )
 
 		switch ( mml->sml_op ) {
 		case LDAP_MOD_DELETE:
+		case SLAP_MOD_SOFTDEL: /* ITS#7487: can be used by syncrepl (in mirror mode?) */
 			if ( vals != NULL ) {
 				for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-					memberof_value_modify( op, rs,
+					memberof_value_modify( op,
 							&vals[ i ], mo->mo_ad_member,
 							&op->o_req_dn, &op->o_req_ndn,
 							NULL, NULL );
@@ -1320,7 +1416,7 @@ memberof_res_modify( Operation *op, SlapReply *rs )
 			op->o_bd->bd_info = (BackendInfo *)on;
 			if ( rc == LDAP_SUCCESS ) {
 				for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-					(void)memberof_value_modify( op, rs,
+					memberof_value_modify( op,
 							&vals[ i ], mo->mo_ad_member,
 							&op->o_req_dn, &op->o_req_ndn,
 							NULL, NULL );
@@ -1334,10 +1430,12 @@ memberof_res_modify( Operation *op, SlapReply *rs )
 			/* fall thru */
 
 		case LDAP_MOD_ADD:
+		case SLAP_MOD_SOFTADD: /* ITS#7487 */
+		case SLAP_MOD_ADD_IF_NOT_PRESENT: /* ITS#7487 */
 			assert( vals != NULL );
 
 			for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-				memberof_value_modify( op, rs,
+				memberof_value_modify( op,
 						&vals[ i ], mo->mo_ad_member,
 						NULL, NULL,
 						&op->o_req_dn, &op->o_req_ndn );
@@ -1358,10 +1456,11 @@ memberof_res_modify( Operation *op, SlapReply *rs )
 
 			switch ( ml->sml_op ) {
 			case LDAP_MOD_DELETE:
+			case SLAP_MOD_SOFTDEL: /* ITS#7487: can be used by syncrepl (in mirror mode?) */
 				vals = ml->sml_nvalues;
 				if ( vals != NULL ) {
 					for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-						memberof_value_modify( op, rs,
+						memberof_value_modify( op,
 								&vals[ i ], mo->mo_ad_memberof,
 								&op->o_req_dn, &op->o_req_ndn,
 								NULL, NULL );
@@ -1376,23 +1475,25 @@ memberof_res_modify( Operation *op, SlapReply *rs )
 				/* delete all ... */
 				if ( vals != NULL ) {
 					for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-						(void)memberof_value_modify( op, rs,
+						memberof_value_modify( op,
 								&vals[ i ], mo->mo_ad_memberof,
 								&op->o_req_dn, &op->o_req_ndn,
 								NULL, NULL );
 					}
 				}
 	
-				if ( ml->sml_op == LDAP_MOD_DELETE || !ml->sml_values ) {
+				if ( ml->sml_op == LDAP_MOD_DELETE || ml->sml_op == SLAP_MOD_SOFTDEL || !ml->sml_values ) {
 					break;
 				}
 				/* fall thru */
 	
 			case LDAP_MOD_ADD:
+			case SLAP_MOD_SOFTADD: /* ITS#7487 */
+			case SLAP_MOD_ADD_IF_NOT_PRESENT : /* ITS#7487 */
 				assert( ml->sml_nvalues != NULL );
 				vals = ml->sml_nvalues;
 				for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-					memberof_value_modify( op, rs,
+					memberof_value_modify( op,
 							&vals[ i ], mo->mo_ad_memberof,
 							NULL, NULL,
 							&op->o_req_dn, &op->o_req_ndn );
@@ -1473,7 +1574,7 @@ memberof_res_modrdn( Operation *op, SlapReply *rs )
 
 		if ( rc == LDAP_SUCCESS ) {
 			for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-				(void)memberof_value_modify( op, rs,
+				memberof_value_modify( op,
 						&vals[ i ], mo->mo_ad_memberof,
 						&op->o_req_dn, &op->o_req_ndn,
 						&newDN, &newNDN );
@@ -1490,7 +1591,7 @@ memberof_res_modrdn( Operation *op, SlapReply *rs )
 
 		if ( rc == LDAP_SUCCESS ) {
 			for ( i = 0; !BER_BVISNULL( &vals[ i ] ); i++ ) {
-				(void)memberof_value_modify( op, rs,
+				memberof_value_modify( op,
 						&vals[ i ], mo->mo_ad_member,
 						&op->o_req_dn, &op->o_req_ndn,
 						&newDN, &newNDN );
@@ -1516,12 +1617,45 @@ memberof_db_init(
 {
 	slap_overinst	*on = (slap_overinst *)be->bd_info;
 	memberof_t		*mo;
+	const char		*text = NULL;
+	int rc;
 
 	mo = (memberof_t *)ch_calloc( 1, sizeof( memberof_t ) );
 
 	/* safe default */
 	mo->mo_dangling_err = LDAP_CONSTRAINT_VIOLATION;
 
+	if ( !ad_memberOf ) {
+		rc = slap_str2ad( SLAPD_MEMBEROF_ATTR, &ad_memberOf, &text );
+		if ( rc != LDAP_SUCCESS ) {
+			Debug( LDAP_DEBUG_ANY, "memberof_db_init: "
+					"unable to find attribute=\"%s\": %s (%d)\n",
+					SLAPD_MEMBEROF_ATTR, text, rc );
+			return rc;
+		}
+	}
+
+	if ( !ad_member ) {
+		rc = slap_str2ad( SLAPD_GROUP_ATTR, &ad_member, &text );
+		if ( rc != LDAP_SUCCESS ) {
+			Debug( LDAP_DEBUG_ANY, "memberof_db_init: "
+					"unable to find attribute=\"%s\": %s (%d)\n",
+					SLAPD_GROUP_ATTR, text, rc );
+			return rc;
+		}
+	}
+
+	if ( !oc_group ) {
+		oc_group = oc_find( SLAPD_GROUP_CLASS );
+		if ( oc_group == NULL ) {
+			Debug( LDAP_DEBUG_ANY,
+					"memberof_db_init: "
+					"unable to find objectClass=\"%s\"\n",
+					SLAPD_GROUP_CLASS );
+			return 1;
+		}
+	}
+
 	on->on_bi.bi_private = (void *)mo;
 
 	return 0;
@@ -1554,9 +1688,10 @@ static ConfigDriver mo_cf_gen;
 
 static ConfigTable mo_cfg[] = {
 	{ "memberof-dn", "modifiersName",
-		2, 2, 0, ARG_MAGIC|ARG_DN|MO_DN, mo_cf_gen,
+		2, 2, 0, ARG_MAGIC|ARG_QUOTE|ARG_DN|MO_DN, mo_cf_gen,
 		"( OLcfgOvAt:18.0 NAME 'olcMemberOfDN' "
 			"DESC 'DN to be used as modifiersName' "
+			"EQUALITY distinguishedNameMatch "
 			"SYNTAX OMsDN SINGLE-VALUE )",
 		NULL, NULL },
 
@@ -1565,6 +1700,7 @@ static ConfigTable mo_cfg[] = {
 		"( OLcfgOvAt:18.1 NAME 'olcMemberOfDangling' "
 			"DESC 'Behavior with respect to dangling members, "
 				"constrained to ignore, drop, error' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString SINGLE-VALUE )",
 		NULL, NULL },
 
@@ -1572,6 +1708,7 @@ static ConfigTable mo_cfg[] = {
 		2, 2, 0, ARG_MAGIC|ARG_ON_OFF|MO_REFINT, mo_cf_gen,
 		"( OLcfgOvAt:18.2 NAME 'olcMemberOfRefInt' "
 			"DESC 'Take care of referential integrity' "
+			"EQUALITY booleanMatch "
 			"SYNTAX OMsBoolean SINGLE-VALUE )",
 		NULL, NULL },
 
@@ -1579,20 +1716,23 @@ static ConfigTable mo_cfg[] = {
 		2, 2, 0, ARG_MAGIC|MO_GROUP_OC, mo_cf_gen,
 		"( OLcfgOvAt:18.3 NAME 'olcMemberOfGroupOC' "
 			"DESC 'Group objectClass' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString SINGLE-VALUE )",
 		NULL, NULL },
 
 	{ "memberof-member-ad", "member attribute",
-		2, 2, 0, ARG_MAGIC|MO_MEMBER_AD, mo_cf_gen,
+		2, 2, 0, ARG_MAGIC|ARG_ATDESC|MO_MEMBER_AD, mo_cf_gen,
 		"( OLcfgOvAt:18.4 NAME 'olcMemberOfMemberAD' "
 			"DESC 'member attribute' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString SINGLE-VALUE )",
 		NULL, NULL },
 
 	{ "memberof-memberof-ad", "memberOf attribute",
-		2, 2, 0, ARG_MAGIC|MO_MEMBER_OF_AD, mo_cf_gen,
+		2, 2, 0, ARG_MAGIC|ARG_ATDESC|MO_MEMBER_OF_AD, mo_cf_gen,
 		"( OLcfgOvAt:18.5 NAME 'olcMemberOfMemberOfAD' "
 			"DESC 'memberOf attribute' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString SINGLE-VALUE )",
 		NULL, NULL },
 
@@ -1610,6 +1750,7 @@ static ConfigTable mo_cfg[] = {
 		2, 2, 0, ARG_MAGIC|MO_DANGLING_ERROR, mo_cf_gen,
 		"( OLcfgOvAt:18.7 NAME 'olcMemberOfDanglingError' "
 			"DESC 'Error code returned in case of dangling back reference' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString SINGLE-VALUE )",
 		NULL, NULL },
 
@@ -1618,7 +1759,7 @@ static ConfigTable mo_cfg[] = {
 
 static ConfigOCs mo_ocs[] = {
 	{ "( OLcfgOvOc:18.1 "
-		"NAME 'olcMemberOf' "
+		"NAME ( 'olcMemberOfConfig' 'olcMemberOf' ) "
 		"DESC 'Member-of configuration' "
 		"SUP olcOverlayConfig "
 		"MAY ( "
@@ -1763,15 +1904,11 @@ mo_cf_gen( ConfigArgs *c )
 			break;
 
 		case MO_MEMBER_AD:
-			if ( mo->mo_ad_member != NULL ){
-				value_add_one( &c->rvalue_vals, &mo->mo_ad_member->ad_cname );
-			}
+			c->value_ad = mo->mo_ad_member;
 			break;
 
 		case MO_MEMBER_OF_AD:
-			if ( mo->mo_ad_memberof != NULL ){
-				value_add_one( &c->rvalue_vals, &mo->mo_ad_memberof->ad_cname );
-			}
+			c->value_ad = mo->mo_ad_memberof;
 			break;
 
 		default:
@@ -1782,7 +1919,52 @@ mo_cf_gen( ConfigArgs *c )
 		return rc;
 
 	} else if ( c->op == LDAP_MOD_DELETE ) {
-		return 1;	/* FIXME */
+		switch( c->type ) {
+		case MO_DN:
+			if ( !BER_BVISNULL( &mo->mo_dn ) ) {
+				ber_memfree( mo->mo_dn.bv_val );
+				ber_memfree( mo->mo_ndn.bv_val );
+				BER_BVZERO( &mo->mo_dn );
+				BER_BVZERO( &mo->mo_ndn );
+			}
+			break;
+
+		case MO_DANGLING:
+			mo->mo_flags &= ~MEMBEROF_FDANGLING_MASK;
+			break;
+
+		case MO_DANGLING_ERROR:
+			mo->mo_dangling_err = LDAP_CONSTRAINT_VIOLATION;
+			break;
+
+		case MO_REFINT:
+			mo->mo_flags &= ~MEMBEROF_FREFINT;
+			break;
+
+#if 0
+		case MO_REVERSE:
+			mo->mo_flags &= ~MEMBEROF_FREVERSE;
+			break;
+#endif
+
+		case MO_GROUP_OC:
+			mo->mo_oc_group = oc_group;
+			memberof_make_group_filter( mo );
+			break;
+
+		case MO_MEMBER_AD:
+			mo->mo_ad_member = ad_member;
+			break;
+
+		case MO_MEMBER_OF_AD:
+			mo->mo_ad_memberof = ad_memberOf;
+			memberof_make_member_filter( mo );
+			break;
+
+		default:
+			assert( 0 );
+			return 1;
+		}
 
 	} else {
 		switch( c->type ) {
@@ -1843,7 +2025,7 @@ mo_cf_gen( ConfigArgs *c )
 					"unable to find group objectClass=\"%s\"",
 					c->argv[ 1 ] );
 				Debug( LDAP_DEBUG_CONFIG, "%s: %s.\n",
-					c->log, c->cr_msg, 0 );
+					c->log, c->cr_msg );
 				return 1;
 			}
 
@@ -1852,19 +2034,7 @@ mo_cf_gen( ConfigArgs *c )
 			} break;
 
 		case MO_MEMBER_AD: {
-			AttributeDescription	*ad = NULL;
-			const char		*text = NULL;
-
-
-			rc = slap_str2ad( c->argv[ 1 ], &ad, &text );
-			if ( rc != LDAP_SUCCESS ) {
-				snprintf( c->cr_msg, sizeof( c->cr_msg ),
-					"unable to find member attribute=\"%s\": %s (%d)",
-					c->argv[ 1 ], text, rc );
-				Debug( LDAP_DEBUG_CONFIG, "%s: %s.\n",
-					c->log, c->cr_msg, 0 );
-				return 1;
-			}
+			AttributeDescription	*ad = c->value_ad;
 
 			if ( !is_at_syntax( ad->ad_type, SLAPD_DN_SYNTAX )		/* e.g. "member" */
 				&& !is_at_syntax( ad->ad_type, SLAPD_NAMEUID_SYNTAX ) )	/* e.g. "uniqueMember" */
@@ -1874,7 +2044,7 @@ mo_cf_gen( ConfigArgs *c )
 					"have DN (%s) or nameUID (%s) syntax",
 					c->argv[ 1 ], SLAPD_DN_SYNTAX, SLAPD_NAMEUID_SYNTAX );
 				Debug( LDAP_DEBUG_CONFIG, "%s: %s.\n",
-					c->log, c->cr_msg, 0 );
+					c->log, c->cr_msg );
 				return 1;
 			}
 
@@ -1882,19 +2052,7 @@ mo_cf_gen( ConfigArgs *c )
 			} break;
 
 		case MO_MEMBER_OF_AD: {
-			AttributeDescription	*ad = NULL;
-			const char		*text = NULL;
-
-
-			rc = slap_str2ad( c->argv[ 1 ], &ad, &text );
-			if ( rc != LDAP_SUCCESS ) {
-				snprintf( c->cr_msg, sizeof( c->cr_msg ),
-					"unable to find memberof attribute=\"%s\": %s (%d)",
-					c->argv[ 1 ], text, rc );
-				Debug( LDAP_DEBUG_CONFIG, "%s: %s.\n",
-					c->log, c->cr_msg, 0 );
-				return 1;
-			}
+			AttributeDescription	*ad = c->value_ad;
 
 			if ( !is_at_syntax( ad->ad_type, SLAPD_DN_SYNTAX )		/* e.g. "member" */
 				&& !is_at_syntax( ad->ad_type, SLAPD_NAMEUID_SYNTAX ) )	/* e.g. "uniqueMember" */
@@ -1904,7 +2062,7 @@ mo_cf_gen( ConfigArgs *c )
 					"have DN (%s) or nameUID (%s) syntax",
 					c->argv[ 1 ], SLAPD_DN_SYNTAX, SLAPD_NAMEUID_SYNTAX );
 				Debug( LDAP_DEBUG_CONFIG, "%s: %s.\n",
-					c->log, c->cr_msg, 0 );
+					c->log, c->cr_msg );
 				return 1;
 			}
 
@@ -1930,37 +2088,17 @@ memberof_db_open(
 	memberof_t	*mo = (memberof_t *)on->on_bi.bi_private;
 	
 	int		rc;
-	const char	*text = NULL;
 
-	if( ! mo->mo_ad_memberof ){
-		rc = slap_str2ad( SLAPD_MEMBEROF_ATTR, &mo->mo_ad_memberof, &text );
-		if ( rc != LDAP_SUCCESS ) {
-			Debug( LDAP_DEBUG_ANY, "memberof_db_open: "
-					"unable to find attribute=\"%s\": %s (%d)\n",
-					SLAPD_MEMBEROF_ATTR, text, rc );
-			return rc;
-		}
+	if ( !mo->mo_ad_memberof ) {
+		mo->mo_ad_memberof = ad_memberOf;
 	}
 
-	if( ! mo->mo_ad_member ){
-		rc = slap_str2ad( SLAPD_GROUP_ATTR, &mo->mo_ad_member, &text );
-		if ( rc != LDAP_SUCCESS ) {
-			Debug( LDAP_DEBUG_ANY, "memberof_db_open: "
-					"unable to find attribute=\"%s\": %s (%d)\n",
-					SLAPD_GROUP_ATTR, text, rc );
-			return rc;
-		}
+	if ( ! mo->mo_ad_member ) {
+		mo->mo_ad_member = ad_member;
 	}
 
-	if( ! mo->mo_oc_group ){
-		mo->mo_oc_group = oc_find( SLAPD_GROUP_CLASS );
-		if ( mo->mo_oc_group == NULL ) {
-			Debug( LDAP_DEBUG_ANY,
-					"memberof_db_open: "
-					"unable to find objectClass=\"%s\"\n",
-					SLAPD_GROUP_CLASS, 0, 0 );
-			return 1;
-		}
+	if ( ! mo->mo_oc_group ) {
+		mo->mo_oc_group = oc_group;
 	}
 
 	if ( BER_BVISNULL( &mo->mo_dn ) && !BER_BVISNULL( &be->be_rootdn ) ) {
@@ -2007,9 +2145,6 @@ memberof_db_destroy(
 	return 0;
 }
 
-/* unused */
-static AttributeDescription	*ad_memberOf;
-
 static struct {
 	char	*desc;
 	AttributeDescription **adp;
@@ -2020,7 +2155,7 @@ static struct {
 		"SYNTAX '1.3.6.1.4.1.1466.115.121.1.12' "
 		"EQUALITY distinguishedNameMatch "	/* added */
 		"USAGE dSAOperation "			/* added; questioned */
-		/* "NO-USER-MODIFICATION " */		/* add? */
+		"NO-USER-MODIFICATION " 		/* added */
 		"X-ORIGIN 'iPlanet Delegated Administrator' )",
 		&ad_memberOf },
 	{ NULL }
@@ -2035,11 +2170,11 @@ memberof_initialize( void )
 	int			code, i;
 
 	for ( i = 0; as[ i ].desc != NULL; i++ ) {
-		code = register_at( as[ i ].desc, as[ i ].adp, 0 );
-		if ( code ) {
+		code = register_at( as[ i ].desc, as[ i ].adp, 1 );
+		if ( code && code != SLAP_SCHERR_ATTR_DUP ) {
 			Debug( LDAP_DEBUG_ANY,
 				"memberof_initialize: register_at #%d failed\n",
-				i, 0, 0 );
+				i );
 			return code;
 		}
 	}
