@@ -2,7 +2,7 @@
 /* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2012 The OpenLDAP Foundation.
+ * Copyright 1999-2022 The OpenLDAP Foundation.
  * Portions Copyright 1999-2003 Howard Chu.
  * Portions Copyright 2000-2003 Pierangelo Masarati.
  * All rights reserved.
@@ -99,9 +99,9 @@ rwm_map_init( struct ldapmap *lm, struct ldapmapping **m )
 	mapping[1].m_src_ad = mapping[0].m_src_ad;
 	mapping[1].m_dst_ad = mapping[1].m_src_ad;
 
-	avl_insert( &lm->map, (caddr_t)&mapping[0], 
+	ldap_avl_insert( &lm->map, (caddr_t)&mapping[0], 
 			rwm_mapping_cmp, rwm_mapping_dup );
-	avl_insert( &lm->remap, (caddr_t)&mapping[1], 
+	ldap_avl_insert( &lm->remap, (caddr_t)&mapping[1], 
 			rwm_mapping_cmp, rwm_mapping_dup );
 
 	*m = mapping;
@@ -138,7 +138,7 @@ rwm_mapping( struct ldapmap *map, struct berval *s, struct ldapmapping **m, int
 	}
 
 	fmapping.m_src = *s;
-	*m = (struct ldapmapping *)avl_find( tree, (caddr_t)&fmapping,
+	*m = (struct ldapmapping *)ldap_avl_find( tree, (caddr_t)&fmapping,
 			rwm_mapping_cmp );
 
 	if ( *m == NULL ) {
@@ -215,7 +215,9 @@ rwm_map_attrnames(
 		return LDAP_NO_MEMORY;
 	}
 
-	for ( i = 0, j = 0; !BER_BVISNULL( &an[i].an_name ); i++ ) {
+	j = 0;
+	if ( an != NULL ) {
+	for ( i = 0; !BER_BVISNULL( &an[i].an_name ); i++ ) {
 		struct ldapmapping	*m;
 		int			at_drop_missing = 0,
 					oc_drop_missing = 0;
@@ -331,6 +333,7 @@ rwm_map_attrnames(
 			}
 		}
 	}
+	}
 
 	if ( op->o_bd->be_extra_anlist != NULL ) {
 		/* we assume be_extra_anlist are already mapped */
@@ -514,7 +517,7 @@ rwm_int_filter_map_rewrite(
 	struct berval		*fstr )
 {
 	int		i;
-	Filter		*p;
+	Filter		*p, ftmp;
 	AttributeDescription *ad;
 	struct berval	atmp,
 			vtmp,
@@ -766,7 +769,7 @@ rwm_int_filter_map_rewrite(
 
 	case -1:
 computed:;
-		filter_free_x( op, f, 0 );
+		f = &ftmp;
 		f->f_choice = SLAPD_FILTER_COMPUTED;
 		f->f_result = SLAPD_COMPARE_UNDEFINED;
 		/* fallthru */
