@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/slapi/slapi_ops.c,v 1.111.2.8 2010/04/13 20:23:50 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2002-2010 The OpenLDAP Foundation.
+ * Copyright 2002-2022 The OpenLDAP Foundation.
  * Portions Copyright 1997,2002-2003 IBM Corporation.
  * All rights reserved.
  *
@@ -225,9 +225,7 @@ slapi_int_connection_init_pb( Slapi_PBlock *pb, ber_tag_t tag )
 	/* should check status of thread calls */
 	ldap_pvt_thread_mutex_init( &conn->c_mutex );
 	ldap_pvt_thread_mutex_init( &conn->c_write1_mutex );
-	ldap_pvt_thread_mutex_init( &conn->c_write2_mutex );
 	ldap_pvt_thread_cond_init( &conn->c_write1_cv );
-	ldap_pvt_thread_cond_init( &conn->c_write2_cv );
 
 	ldap_pvt_thread_mutex_lock( &conn->c_mutex );
 
@@ -235,6 +233,7 @@ slapi_int_connection_init_pb( Slapi_PBlock *pb, ber_tag_t tag )
 	conn->c_n_ops_executing = 0;
 	conn->c_n_ops_pending = 0;
 	conn->c_n_ops_completed = 0;
+	conn->c_n_ops_async = 0;
 
 	conn->c_n_get = 0;
 	conn->c_n_read = 0;
@@ -252,8 +251,7 @@ slapi_int_connection_init_pb( Slapi_PBlock *pb, ber_tag_t tag )
 	 */
 	connection_assign_nextid( conn );
 
-	conn->c_conn_state  = 0x01;	/* SLAP_C_ACTIVE */
-	conn->c_struct_state = 0x02;	/* SLAP_C_USED */
+	conn->c_conn_state  = SLAP_C_ACTIVE;
 
 	conn->c_ssf = conn->c_transport_ssf = local_ssf;
 	conn->c_tls_ssf = 0;
@@ -384,7 +382,6 @@ slapi_int_connection_done_pb( Slapi_PBlock *pb )
 static int
 slapi_int_func_internal_pb( Slapi_PBlock *pb, slap_operation_t which )
 {
-	BI_op_bind		**func;
 	SlapReply		*rs = pb->pb_rs;
 	int			rc;
 
@@ -397,9 +394,7 @@ slapi_int_func_internal_pb( Slapi_PBlock *pb, slap_operation_t which )
 	}
 
 	pb->pb_op->o_bd = frontendDB;
-	func = &frontendDB->be_bind;
-
-	return func[which]( pb->pb_op, pb->pb_rs );
+	return (&frontendDB->be_bind)[which]( pb->pb_op, pb->pb_rs );
 }
 
 int
@@ -953,4 +948,3 @@ slapi_delete_internal(
 }
 
 #endif /* LDAP_SLAPI */
-
