@@ -1,8 +1,8 @@
 /* syntax.c - routines to manage syntax definitions */
-/* $OpenLDAP: pkg/ldap/servers/slapd/syntax.c,v 1.43.2.7 2010/04/13 20:23:22 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2010 The OpenLDAP Foundation.
+ * Copyright 1998-2022 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -60,7 +60,7 @@ syn_find( const char *synname )
 {
 	struct sindexrec	*sir = NULL;
 
-	if ( (sir = avl_find( syn_index, synname, syn_index_name_cmp )) != NULL ) {
+	if ( (sir = ldap_avl_find( syn_index, synname, syn_index_name_cmp )) != NULL ) {
 		return( sir->sir_syn );
 	}
 	return( NULL );
@@ -113,7 +113,7 @@ syn_destroy( void )
 {
 	Syntax	*s;
 
-	avl_free( syn_index, ldap_memfree );
+	ldap_avl_free( syn_index, ldap_memfree );
 	while( !LDAP_STAILQ_EMPTY( &syn_list ) ) {
 		s = LDAP_STAILQ_FIRST( &syn_list );
 		LDAP_STAILQ_REMOVE_HEAD( &syn_list, ssyn_next );
@@ -138,13 +138,13 @@ syn_insert(
 		sir = (struct sindexrec *)
 			SLAP_CALLOC( 1, sizeof(struct sindexrec) );
 		if( sir == NULL ) {
-			Debug( LDAP_DEBUG_ANY, "SLAP_CALLOC Error\n", 0, 0, 0 );
+			Debug( LDAP_DEBUG_ANY, "SLAP_CALLOC Error\n" );
 			return LDAP_OTHER;
 		}
 		sir->sir_name = ssyn->ssyn_oid;
 		sir->sir_syn = ssyn;
-		if ( avl_insert( &syn_index, (caddr_t) sir,
-		                 syn_index_cmp, avl_dup_error ) ) {
+		if ( ldap_avl_insert( &syn_index, (caddr_t) sir,
+		                 syn_index_cmp, ldap_avl_dup_error ) ) {
 			*err = ssyn->ssyn_oid;
 			ldap_memfree(sir);
 			return SLAP_SCHERR_SYN_DUP;
@@ -184,7 +184,7 @@ syn_add(
 
 	ssyn = (Syntax *) SLAP_CALLOC( 1, sizeof(Syntax) );
 	if ( ssyn == NULL ) {
-		Debug( LDAP_DEBUG_ANY, "SLAP_CALLOC Error\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_ANY, "SLAP_CALLOC Error\n" );
 		return SLAP_SCHERR_OUTOFMEM;
 	}
 
@@ -219,18 +219,20 @@ syn_add(
 			}
 
 			assert( (*lsei)->lsei_values != NULL );
-			if ( (*lsei)->lsei_values[0] == '\0'
-				|| (*lsei)->lsei_values[1] != '\0' )
+			if ( (*lsei)->lsei_values[0] == NULL
+				|| (*lsei)->lsei_values[1] != NULL )
 			{
 				Debug( LDAP_DEBUG_ANY, "syn_add(%s): exactly one substitute syntax must be present\n",
-					ssyn->ssyn_syn.syn_oid, 0, 0 );
+					ssyn->ssyn_syn.syn_oid );
+				SLAP_FREE( ssyn );
 				return SLAP_SCHERR_SYN_SUBST_NOT_SPECIFIED;
 			}
 
 			subst = syn_find( (*lsei)->lsei_values[0] );
 			if ( subst == NULL ) {
 				Debug( LDAP_DEBUG_ANY, "syn_add(%s): substitute syntax %s not found\n",
-					ssyn->ssyn_syn.syn_oid, (*lsei)->lsei_values[0], 0 );
+					ssyn->ssyn_syn.syn_oid, (*lsei)->lsei_values[0] );
+				SLAP_FREE( ssyn );
 				return SLAP_SCHERR_SYN_SUBST_NOT_FOUND;
 			}
 			break;
@@ -259,7 +261,7 @@ syn_add(
 		ssyn->ssyn_sups = (Syntax **)SLAP_CALLOC( cnt + 1,
 			sizeof( Syntax * ) );
 		if ( ssyn->ssyn_sups == NULL ) {
-			Debug( LDAP_DEBUG_ANY, "SLAP_CALLOC Error\n", 0, 0, 0 );
+			Debug( LDAP_DEBUG_ANY, "SLAP_CALLOC Error\n" );
 			code = SLAP_SCHERR_OUTOFMEM;
 
 		} else {
@@ -349,7 +351,7 @@ syn_schema_info( Entry *e )
 		}
 #if 0
 		Debug( LDAP_DEBUG_TRACE, "Merging syn [%ld] %s\n",
-	       (long) val.bv_len, val.bv_val, 0 );
+	       (long) val.bv_len, val.bv_val );
 #endif
 
 		nval.bv_val = syn->ssyn_oid;
