@@ -1,8 +1,8 @@
 /* user.c - set user id, group id and group access list */
-/* $OpenLDAP: pkg/ldap/servers/slapd/user.c,v 1.25.2.5 2010/04/13 20:23:22 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2010 The OpenLDAP Foundation.
+ * Copyright 1998-2022 The OpenLDAP Foundation.
  * Portions Copyright 1999 PM Lashley.
  * All rights reserved.
  *
@@ -38,7 +38,6 @@
 
 /*
  * Set real and effective user id and group id, and group access list
- * The user and group arguments are freed.
  */
 
 void
@@ -56,7 +55,7 @@ slap_init_user( char *user, char *group )
 	    got_uid = 1;
 	    if ( lutil_atou( &u, user ) != 0 ) {
 		Debug( LDAP_DEBUG_ANY, "Unble to parse user %s\n",
-		       user, 0, 0 );
+		       user );
 
 		exit( EXIT_FAILURE );
 	    }
@@ -65,7 +64,6 @@ slap_init_user( char *user, char *group )
 	    pwd = getpwuid( uid );
 	    goto did_getpw;
 #else
-	    free( user );
 	    user = NULL;
 #endif
 	} else {
@@ -73,13 +71,12 @@ slap_init_user( char *user, char *group )
 	did_getpw:
 	    if ( pwd == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "No passwd entry for user %s\n",
-		       user, 0, 0 );
+		       user );
 
 		exit( EXIT_FAILURE );
 	    }
 	    if ( got_uid ) {
-		free( user );
-		user = (pwd != NULL ? ch_strdup( pwd->pw_name ) : NULL);
+		user = (pwd != NULL ? pwd->pw_name : NULL);
 	    } else {
 		got_uid = 1;
 		uid = pwd->pw_uid;
@@ -99,7 +96,7 @@ slap_init_user( char *user, char *group )
 
 	    if ( lutil_atou( &g, group ) != 0 ) {
 		Debug( LDAP_DEBUG_ANY, "Unble to parse group %s\n",
-		       group, 0, 0 );
+		       group );
 
 		exit( EXIT_FAILURE );
 	    }
@@ -115,23 +112,21 @@ slap_init_user( char *user, char *group )
 	did_group:
 	    if ( grp == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "No group entry for group %s\n",
-		       group, 0, 0 );
+		       group );
 
 		exit( EXIT_FAILURE );
 	    }
 	}
-	free( group );
 	got_gid = 1;
     }
 
     if ( user ) {
 	if ( getuid() == 0 && initgroups( user, gid ) != 0 ) {
 	    Debug( LDAP_DEBUG_ANY,
-		   "Could not set the group access (gid) list\n", 0, 0, 0 );
+		   "Could not set the group access (gid) list\n" );
 
 	    exit( EXIT_FAILURE );
 	}
-	free( user );
     }
 
 #ifdef HAVE_ENDGRENT
@@ -141,14 +136,14 @@ slap_init_user( char *user, char *group )
     if ( got_gid ) {
 	if ( setgid( gid ) != 0 ) {
 	    Debug( LDAP_DEBUG_ANY, "Could not set real group id to %d\n",
-		       (int) gid, 0, 0 );
+		       (int) gid );
 
 	    exit( EXIT_FAILURE );
 	}
 #ifdef HAVE_SETEGID
 	if ( setegid( gid ) != 0 ) {
 	    Debug( LDAP_DEBUG_ANY, "Could not set effective group id to %d\n",
-		       (int) gid, 0, 0 );
+		       (int) gid );
 
 	    exit( EXIT_FAILURE );
 	}
@@ -158,14 +153,14 @@ slap_init_user( char *user, char *group )
     if ( got_uid ) {
 	if ( setuid( uid ) != 0 ) {
 	    Debug( LDAP_DEBUG_ANY, "Could not set real user id to %d\n",
-		       (int) uid, 0, 0 );
+		       (int) uid );
 
 	    exit( EXIT_FAILURE );
 	}
 #ifdef HAVE_SETEUID
 	if ( seteuid( uid ) != 0 ) {
 	    Debug( LDAP_DEBUG_ANY, "Could not set effective user id to %d\n",
-		       (int) uid, 0, 0 );
+		       (int) uid );
 
 	    exit( EXIT_FAILURE );
 	}
